"""
OXASL: Multi-step Bayesian model fitting for ASL data

Based on the BASIL shell command from FSL
"""
import sys
import math

import numpy as np
import scipy.ndimage

from fsl.wrappers import LOAD
from fsl.data.image import Image

from oxasl import __version__, __timestamp__, reg

def run(wsp, prefit=True, **kwargs):
    """
    Run multi-step model fitting on ASL data in a workspace

    :param wsp: Workspace object
    :param prefit: If True, run a pre-fitting step using the mean over repeats of the ASL data

    Required workspace attributes
    -----------------------------

     - ``asldata`` : AslImage object

    Optional workspace attributes
    -----------------------------

     - ``mask`` : Brain mask (fsl.Image)
     - ``wp`` : If True, use 'white paper' mode (Alsop et al) - modifies some defaults and infers tissue component only
     - ``infertiss`` : If True, infer tissue component (default: True)
     - ``inferbat`` : If True, infer bolus arrival time (default: False)
     - ``infertau`` : If True, infer bolus duration (default: False)
     - ``inferart`` : If True, infer arterial component (default: False)
     - ``infert1`` : If True, infer T1 (default: False)
     - ``inferpc`` : If True, infer PC (default: False)
     - ``t1``: Assumed/initial estimate for tissue T1 (default: 1.65 in white paper mode, 1.3 otherwise)
     - ``t1b``: Assumed/initial estimate for blood T1 (default: 1.65)
     - ``bat``: Assumed/initial estimate for bolus arrival time (s) (default 0 in white paper mode, 1.3 for CASL, 0.7 otherwise)
     - ``t1im`` : T1 map as Image
     - ``pgm`` :  Grey matter partial volume map as Image
     - ``pwm`` : White matter partial volume map as Image
     - ``initmvn`` : MVN structure to use as initialization as Image
     - ``spatial`` : If True, include final spatial VB step (default: False)
     - ``onestep`` : If True, do all inference in a single step (default: False)
     - ``basil_options`` : Optional dictionary of additional options for underlying model
    """
    wsp.log.write("\nRunning BASIL Bayesian modelling on ASL data in '%s' data space\n" % wsp.ifnone("image_space", "asl"))

    # Set default options based on data characteristics
    _set_defaults(wsp)
    _define_mask(wsp)

    # Do pre-fit on averaged data if requested
    if prefit and max(wsp.asldata.rpts) > 1:
        wsp.log.write(" - Doing initial fit on mean at each TI\n\n")
        init_wsp = wsp.sub("init")
        main_wsp = wsp.sub("main")
        run_multistep_fitting(init_wsp, wsp.asldata.mean_across_repeats())
        wsp.basil_options["continue-from-mvn"] = wsp.init.finalstep.finalMVN
        main_wsp.initmvn = wsp.basil_options["continue-from-mvn"]
    else:
        main_wsp = wsp

    # Then do main run on full ASL data
    wsp.log.write("\n - Doing fit on full ASL data\n\n")
    run_multistep_fitting(main_wsp, wsp.asldata)
    wsp.finalstep = main_wsp.finalstep

def _set_defaults(wsp):
    # Single or Multi TI setup
    if wsp.asldata.ntis == 1:
        # Single TI data - don't try to infer arterial component of bolus duration, we don't have enough info
        wsp.log.write(" - Operating in Single TI mode - no arterial component, fixed bolus duration\n")
        wsp.inferart = False
        wsp.infertau = False
        batsd_default = 0.1
    else:
        # For multi TI/PLD data, set a more liberal prior for tissue ATT since we should be able to
        # determine this from the data. NB this leaves the arterial BAT alone.
        batsd_default = 1

    if wsp.wp:
        # White paper mode - this overrides defaults, but can be overwritten by command line
        # specification of individual parameters
        wsp.log.write(" - Analysis in white paper mode: T1 default=1.65, BAT default=0, voxelwise calibration\n")
        t1_default = 1.65
        bat_default = 0.0
    else:
        t1_default = 1.3
        if wsp.asldata.casl:
            bat_default = 1.3
        else:
            bat_default = 0.7

    if wsp.t1 is None:
        wsp.t1 = t1_default
    if wsp.t1b is None:
        wsp.t1b = 1.65
    if wsp.bat is None:
        wsp.bat = bat_default
    if wsp.batsd is None:
        wsp.batsd = batsd_default
    if wsp.infertiss is None:
        wsp.infertiss = True

    # if we are doing CASL then fix the bolus duration, unless explicitly told us otherwise
    if wsp.infertau is None:
        wsp.infertau = not wsp.asldata.casl

    # Take a copy of extra user-specified BASIL options
    wsp.basil_options = dict(wsp.ifnone("basil_options", {}))

    # If we only have one volume, set a nominal noise prior as it is not possible to
    # estimate from the data
    if wsp.asldata.nvols / wsp.asldata.ntc == 1:
        wsp.log.write(" - Restricting noise prior as only one ASL volume\n")
        wsp.basil_options["prior-noise-stddev"] = 1.0

def _define_mask(wsp):
    # Save a copy of the *analysis* mask to be used for this data as this might change later
    # e.g. for PVC. Fitting may occur in a different mask. Two locations for compatibility...
    if wsp.rois is not None and wsp.rois.mask is not None:
        wsp.analysis_mask = wsp.rois.mask
    else:
        wsp.analysis_mask = wsp.mask

    # Now determine the *fitting* mask
    mask_policy = wsp.ifnone("basil_mask", "default")
    if mask_policy in ("default", "dilated"):
        wsp.log.write(" - Using pipeline analysis mask\n")
        # Two possible locations for compatibility
        if wsp.rois is not None and wsp.rois.mask is not None:
            wsp.basil_mask = wsp.rois.mask
        else:
            wsp.basil_mask = wsp.mask
        if mask_policy == "dilated":
            # Use 3x3x3 kernel for compatibility with fslmaths default
            wsp.log.write(" - Dilating mask for Basil analysis\n")
            struct = scipy.ndimage.generate_binary_structure(3, 3)
            wsp.basil_mask = Image(scipy.ndimage.binary_dilation(wsp.basil_mask.data, structure=struct).astype(np.int32), header=wsp.basil_mask.header)
    elif mask_policy == "none":
            wsp.log.write(" - Not using mask for Basil - will fit every voxel\n")
            wsp.basil_mask = Image(np.ones(wsp.asldata.data.shape[:3]), header=wsp.asldata.header)
    else:
        raise ValueError("Unrecognized mask policy: %s" % mask_policy)

def run_multistep_fitting(wsp, asldata):
    """
    Run multi-step Bayesian model fitting on ASL data

    See ``run`` for details of workspace attributes used

    :param wsp: Workspace object
    :param asldata: AslImage object to use as input data
    """
    if len(asldata.tes) > 1:
        steps = fitting_steps_multite(wsp, asldata)
    else:
        steps = fitting_steps(wsp, asldata)

    prev_result = None
    wsp.asldata_diff = asldata.diff().reorder("rt")

    for idx, step in enumerate(steps):
        step_wsp = wsp.sub("step%i" % (idx+1))
        desc = "Step %i of %i: %s" % (idx+1, len(steps), step.desc)
        if prev_result is not None:
            desc += " - Initialise with step %i" % idx
        step_wsp.log.write(desc + "     ")
        result = step.run(prev_result, log=wsp.log, fsllog=wsp.fsllog)
        for key, value in result.items():
            if key == "modelfit":
                # Treat model fit specially - make it an AslImage and also output a mean
                # across repeats version for comparison
                value = wsp.asldata_diff.derived(value.data, header=value.header)
                modelfit_mean = value.mean_across_repeats()
                setattr(step_wsp, "modelfit_mean", modelfit_mean)
            setattr(step_wsp, key, value)

        if step_wsp.logfile is not None and step_wsp.savedir is not None:
            step_wsp.set_item("logfile", step_wsp.logfile, save_fn=str)

        prev_result = result
    wsp.finalstep = step_wsp
    wsp.log.write("\nEnd\n")

def _handle_slicedt(wsp, options):
    """
    Handle the slicedt parameter if we are quantifying in
    a space other than the native ASL space

    Since slicedt is defined relative to the acquisition (ASL) space we
    generate a 3D volume in that space where every voxel contains the
    effective slice time offset. We then transform this into the 
    quantification space we are using and pass it to model fitting using the 'tiimg'
    option.
    """
    img_space = wsp.ifnone("image_space", "asl")
    if img_space != "asl":
        asldata = options["data"]
        _x, _y, z, _t = np.indices(list(asldata.data.shape[:3]) + [asldata.ntis,])
        tis_arr = np.array(asldata.tis) + (z.astype(np.float32) * options["slicedt"])
        tis_img = Image(tis_arr, header=options["data"].header)
        wsp.tiimg = reg.change_space(wsp, tis_img, wsp.ifnone("image_space", "asl"))

        del options["slicedt"]
        ti_idx = 1
        while "ti%i" % ti_idx in options:
            del options["ti%i" % ti_idx]
            ti_idx += 1

        options["tiimg"] = wsp.tiimg

def fitting_steps(wsp, asldata):
    """
    Get the steps required for a BASIL run

    This is separated for the case where an alternative process wants to run
    the actual modelling, or so that the steps can be checked prior to doing
    an actual run.

    Arguments are the same as the ``basil`` function. No workspace is required.
    """
    if asldata is None:
        raise ValueError("Input ASL data is None")

    wsp.log.write("BASIL v%s\n" % __version__)
    asldata.summary(log=wsp.log)
    asldata = asldata.diff().reorder("rt")

    # Default model fitting options for VB runs and spatial steps. Note that values
    # could are None (e.g. sliceband) - these will not be passed to the model fitting step
    options = {
        "data" : asldata,
        "model" : "aslrest",
        "disp" : "none",
        "exch" : "mix",
        "method" : "vb",
        "noise" : "white",
        "allow-bad-voxels" : True,
        "max-iterations" : 20,
        "convergence" : "trialmode",
        "max-trials" : 10,
        "save-mean" : True,
        "save-mvn" : True,
        "save-std" : True,
        "save-noise-mean" : True,
        "save-model-fit" : True,
        "save-residuals" : wsp.ifnone("output_residuals", False),
    }

    if wsp.basil_mask is not None:
        options["mask"] = wsp.basil_mask

    # We choose to pass TIs (not PLDs). The asldata object ensures that
    # TIs are correctly derived from PLDs, when these are specified, by adding
    # the bolus duration.
    options["tis"] = list(asldata.tis)
    options["rpts"] = list(asldata.rpts)

    # Bolus duration - use a single value where possible as cannot infer otherwise
    taus = getattr(asldata, "taus", [1.8,])
    if min(taus) == max(taus):
        options["tau"] = taus[0]
    else:
        options["taus"] = list(taus)

    # Other asl data parameters
    for attr in ("casl", "slicedt", "sliceband"):
        if getattr(asldata, attr, None) is not None:
            options[attr] = getattr(asldata, attr)

    _handle_slicedt(wsp, options)

    if wsp.noiseprior:
        # Use an informative noise prior
        if wsp.noisesd is None:
            snr = wsp.ifnone("snr", 10)
            wsp.log.write(" - Using SNR of %f to set noise std dev\n" % snr)

            # Estimate signal magntiude FIXME diffdata_mean is always 3D?
            if wsp.diffdata_mean.ndim > 3:
                datamax = np.amax(wsp.diffdata_mean.data, 3)
            else:
                datamax = wsp.diffdata_mean.data
            brain_mag = np.mean(datamax.data[wsp.basil_mask.data != 0])
            # this will correspond to whole brain CBF (roughly) - about 0.5 of GM
            noisesd = math.sqrt(brain_mag * 2 / snr)
        else:
            noisesd = wsp.noisesd
        wsp.log.write(" - Using a prior noise sd of: %f\n" % noisesd)
        options["prior-noise-stddev"] = noisesd

    # Add Basil-specific options defined on the workspace
    options.update(wsp.ifnone("basil_options", {}))

    # Additional optional workspace arguments
    for attr in ("t1", "t1b", "bat", "FA", "pwm", "pgm", "batsd"):
        value = getattr(wsp, attr)
        if value is not None:
            options[attr] = value

    # Options for final spatial step
    prior_type_spatial = "M"
    prior_type_mvs = "A"
    options_svb = {
        "method" : "spatialvb",
        "param-spatial-priors" : "N+",
        "convergence" : "maxits",
        "max-iterations": 20,
    }

    wsp.log.write("Model: %s\n" % options["model"])
    wsp.log.write("Dispersion model: %s\n" % options["disp"])
    wsp.log.write("Compartment exchange model: %s\n" % options["exch"])
    inferdisp = options["disp"] != "none"
    inferexch = options["exch"] != "mix"

    # Partial volume correction
    pvcorr = "pgm" in options or "pwm" in options
    if pvcorr:
        if not wsp.infertiss:
            raise ValueError("ERROR: PV correction is not compatible with --artonly option (there is no tissue component)")

        options["incpve"] = True
        if "pgm" not in options or "pwm" not in options:
            raise ValueError("Only one partial volume map (GM / WM) was supplied for PV correctioN")
        # Need a spatial step with more iterations for the PV correction
        wsp.spatial = True
        options_svb["max-iterations"] = 200
        # Ignore partial volumes below 0.1
        pgm_img = options.pop("pgm")
        pwm_img = options.pop("pwm")
        pgm = np.copy(pgm_img.data)
        pwm = np.copy(pwm_img.data)
        pgm[pgm < 0.1] = 0
        pgm[pgm > 1] = 1
        pwm[pwm < 0.1] = 0
        pwm[pwm > 1] = 1
        pgm = Image(pgm, header=pgm_img.header)
        pwm = Image(pwm, header=pwm_img.header)

    # Set general parameter inference and inclusion
    if wsp.infertiss:
        options["inctiss"] = True
    if wsp.inferbat:
        options["incbat"] = True
        options["inferbat"] = True # Infer in first step
    if wsp.inferart:
        options["incart"] = True
    if wsp.inferpc:
        options["incpc"] = True
    if wsp.infertau:
        options["inctau"] = True
    if wsp.infert1:
        options["inct1"] = True

    # Keep track of the number of spatial priors specified by name
    spriors = 1

    if wsp.initmvn:
        # we are being supplied with an initial MVN
        wsp.log.write("Initial MVN being loaded %s\n" % wsp.initmvn.name)
        options["continue-from-mvn"] = wsp.initmvn

    # T1 image prior
    if wsp.t1im is not None:
        spriors = _add_prior(options, spriors, "T_1", type="I", image=wsp.t1im)

    # BAT image prior
    if wsp.batim is not None:
        # With a BAT image prior we must include BAT even if we are not inferring it
        # (in this case the image prior will be treated as ground truth)
        spriors = _add_prior(options, spriors, "delttiss", type="I", image=wsp.batim)
        options["incbat"] = True

    steps = []
    components = ""

    ### --- TISSUE MODULE ---
    if wsp.infertiss:
        components += " Tissue "
        options["infertiss"] = True
        step_desc = "VB - %s" % components
        if not wsp.onestep:
            steps.append(FittingStep(wsp, options, step_desc))

        # setup spatial priors ready
        spriors = _add_prior(options_svb, spriors, "ftiss", type=prior_type_spatial)

    ### --- ARTERIAL MODULE ---
    if wsp.inferart:
        components += " Arterial "
        options["inferart"] = True
        step_desc = "VB - %s" % components
        if not wsp.onestep:
            steps.append(FittingStep(wsp, options, step_desc))

        # setup spatial priors ready
        spriors = _add_prior(options_svb, spriors, "fblood", type=prior_type_mvs)

    ### --- BOLUS DURATION MODULE ---
    if wsp.infertau:
        components += " Bolus duration "
        options["infertau"] = True
        step_desc = "VB - %s" % components
        if not wsp.onestep:
            steps.append(FittingStep(wsp, options, step_desc))

    ### --- MODEL EXTENSIONS MODULE ---
    # Add variable dispersion and/or exchange parameters and/or pre-capiliary
    if inferdisp or inferexch or wsp.inferpc:
        if inferdisp:
            components += " dispersion"
            options["inferdisp"] = True
        if inferexch:
            components += " exchange"
            options["inferexch"] = True
        if wsp.inferpc:
            components += " pre-capiliary"
            options["inferpc"] = True

        step_desc = "VB - %s" % components
        if not wsp.onestep:
            steps.append(FittingStep(wsp, options, step_desc))

    ### --- T1 MODULE ---
    if wsp.infert1:
        components += " T1 "
        options["infert1"] = True
        step_desc = "VB - %s" % components
        if not wsp.onestep:
            steps.append(FittingStep(wsp, options, step_desc))

    ### --- PV CORRECTION MODULE ---
    if pvcorr:
        # setup ready for PV correction, which has to be done with spatial priors
        components += " PVE"
        options["pvcorr"] = True

        # set the image priors for the PV maps
        spriors = _add_prior(options, spriors, "pvgm", type="I", image=pgm)
        spriors = _add_prior(options, spriors, "pvwm", type="I", image=pwm)
        spriors = _add_prior(options, spriors, "fwm", type="M")

        if steps:
            # Add initialisaiton step for PV correction - ONLY if we have something to init from
            steps.append(PvcInitStep(wsp, {"data" : asldata, "mask" : wsp.basil_mask, "pgm" : pgm, "pwm" : pwm}, "PVC initialisation"))

    ### --- SPATIAL MODULE ---
    if wsp.spatial:
        step_desc = "Spatial VB - %s" % components
        options.update(options_svb)
        del options["max-trials"]

        if not wsp.onestep:
            steps.append(FittingStep(wsp, options, step_desc))

    ### --- SINGLE-STEP OPTION ---
    if wsp.onestep:
        steps.append(FittingStep(wsp, options, step_desc))

    if not steps:
        raise ValueError("No steps were generated - no parameters were set to be inferred")

    return steps

def fitting_steps_multite(wsp, asldata, **kwargs):
    """
    Get the steps required for a BASIL run on multi-TE data

    This is separated for the case where an alternative process wants to run
    the actual modelling, or so that the steps can be checked prior to doing
    an actual run.

    Arguments are the same as the ``fitting_steps`` function.
    """
    if asldata is None:
        raise ValueError("Input ASL data is None")

    wsp.log.write("BASIL v%s\n" % __version__)
    asldata.summary(log=wsp.log)
    asldata = asldata.diff().reorder("rt")

    # Default model fitting options for VB runs and spatial steps.
    options = {
        "data" : asldata,
        "model" : "asl_multite",
        "method" : "vb",
        "noise" : "white",
        "allow-bad-voxels" : True,
        "max-iterations" : 20,
        "convergence" : "trialmode",
        "max-trials" : 10,
        "save-mean" : True,
        "save-mvn" : True,
        "save-std" : True,
        "save-model-fit" : True,
        "mask" : wsp.basil_mask,
    }

    # We choose to pass TIs (not PLDs). The asldata object ensures that
    # TIs are correctly derived from PLDs, when these are specified, by adding
    # the bolus duration.
    _list_option(options, asldata.tis, "ti")

    # Pass multiple TEs
    _list_option(options, asldata.tes, "te")

    # Bolus duration must be constant for multi-TE model
    if min(asldata.taus) != max(asldata.taus):
        raise ValueError("Multi-TE model does not support variable bolus durations")
    else:
        options["tau"] = asldata.taus[0]

    # Repeats must be constant for multi-TE model
    if min(asldata.rpts) != max(asldata.rpts):
        raise ValueError("Multi-TE model does not support variable repeats")
    else:
        options["repeats"] = asldata.rpts[0]

    # Other asl data parameters
    for attr in ("casl", "slicedt", "sliceband"):
        if getattr(asldata, attr, None) is not None:
            options[attr] = getattr(asldata, attr)

    # Keyword arguments override options
    options.update(kwargs)

    # Additional optional workspace arguments
    for attr in ("t1", "t1b", "t2", "t2b"):
        value = getattr(wsp, attr)
        if value is not None:
            if attr.startswith("t2"):
                # Model expects T2 in seconds not ms
                options[attr] = float(value) / 1000
            else:
                options[attr] = value

    # Options for final spatial step
    prior_type_spatial = "M"
    prior_type_mvs = "A"
    options_svb = {
        "method" : "spatialvb",
        "param-spatial-priors" : "N+",
        "convergence" : "maxits",
        "max-iterations": 20,
    }

    wsp.log.write("Model: %s\n" % options["model"])
    
    # Set general parameter inference and inclusion
    if not wsp.infertiss:
        wsp.log.write("WARNING: infertiss=False but ftiss is always inferred in multi-TE model\n")
    if not wsp.inferbat:
        wsp.log.write("WARNING: inferbat=False but BAT is always inferred in multi-TE model\n")
    if wsp.inferart:
        wsp.log.write("WARNING: inferart=True but multi-TE model does not support arterial component\n")
    if wsp.infertau:
        options["infertau"] = True
    if wsp.infert1:
        options["infert1"] = True
    if wsp.infert2:
        options["infert2"] = True

    # Keep track of the number of spatial priors specified by name
    spriors = 1

    if wsp.initmvn:
        # we are being supplied with an initial MVN
        wsp.log.write("Initial MVN being loaded %s\n" % wsp.initmvn.name)
        options["continue-from-mvn"] = wsp.initmvn

    # T1 image prior
    if wsp.t1im:
        spriors = _add_prior(options, spriors, "T_1", type="I", image=wsp.t1im)

    # BAT image prior
    if wsp.batim is not None:
        # With a BAT image prior we must include BAT even if we are not inferring it
        # (in this case the image prior will be treated as ground truth)
        spriors = _add_prior(options, spriors, "delttiss", type="I", image=wsp.batim)
        options["incbat"] = True

    steps = []
    components = ""

    ### --- TISSUE MODULE ---
    #if wsp.infertiss:
    if True:
        components += " Tissue"

        ### Inference options
        if wsp.infertau:
            components += " Bolus duration"
            options["infertau"] = True
        if wsp.infert1:
            components += " T1"
            options["infert1"] = True
        if wsp.infertexch:
            components += " Exchange time"
            options["infertexch"] = True

        step_desc = "VB - %s" % components
        if not wsp.onestep:
            steps.append(FittingStep(wsp, options, step_desc))

        # Setup spatial priors ready
        spriors = _add_prior(options_svb, spriors, "ftiss", type=prior_type_spatial)

    ### --- SPATIAL MODULE ---
    if wsp.spatial:
        step_desc = "Spatial VB - %s" % components
        options.update(options_svb)
        del options["max-trials"]

        if not wsp.onestep:
            steps.append(FittingStep(wsp, options, step_desc))

    ### --- SINGLE-STEP OPTION ---
    if wsp.onestep:
        steps.append(FittingStep(wsp, options, step_desc))

    if not steps:
        raise ValueError("No steps were generated - no parameters were set to be inferred")

    return steps

def _list_option(options, values, name):
    for idx, value in enumerate(values):
        options["%s%i" % (name, idx+1)] = value

def _add_prior(options, prior_idx, param, **kwargs):
    options["PSP_byname%i" % prior_idx] = param
    for key, value in kwargs.items():
        options["PSP_byname%i_%s" % (prior_idx, key)] = value
    return prior_idx + 1

class Step(object):
    """
    A step in the fitting process
    """
    def __init__(self, wsp, options, desc):
        self.options = dict(options)
        self.desc = desc
        # Need to convert all images to quantification image space
        for key in list(options.keys()):
            poss_img = self.options[key]
            if isinstance(poss_img, Image):
                image_space = wsp.ifnone("image_space", "asl")
                self.options[key] = reg.change_space(wsp, poss_img, image_space, mask=(key == 'mask'))

class FittingStep(Step):
    """
    A step which fits a model to the data
    """
    def __init__(self, wsp, options, desc):
        Step.__init__(self, wsp, options, desc)
        self._impl = wsp.ifnone("basil_method", "fabber")

    def run(self, prev_output, log=sys.stdout, fsllog=None, **kwargs):
        """
        Run model fitting (fabber or vaby), initialising it from the output of a previous step
        """
        if prev_output is not None:
            self.options["continue-from-mvn"] = prev_output["finalMVN"]
        if self._impl == "fabber":
            from oxasl.wrappers import fabber
            fabber_options = dict(self.options)
            for k, v in self.options.items():
                if isinstance(v, list):
                    fabber_options.pop(k)
                    _list_option(fabber_options, v, k[:-1])
            ret = fabber(fabber_options, output=LOAD, progress_log=log, log=fsllog, **kwargs)
        elif self._impl == "vaby":
            import vaby
            vaby_options = {}
            for k, v in self.options.items():
                k = k.replace("-", "_")
                if isinstance(v, Image):
                    v = v.nibImage
                vaby_options[k] = v
            vaby_options["model_name"] = vaby_options.pop("model")
            vaby_options["initial_posterior"] = vaby_options.pop("continue-from-mvn", None)
            vaby_options["repeats"] = vaby_options.pop("rpts", None)
            vaby_options["method"] = "avb"
            vaby_options["save_posterior"] = True
            output = {}
            _runtime, _state = vaby.run(outdict=output, **vaby_options)
            ret = {}
            for k, v in output.items():
                v = Image(v)
                if k == "posterior_native":
                    ret["finalMVN"] = v
                else:
                    ret[k.replace("_native", "")] = v
        else:
            raise ValueError("Unrecognized model fitting implementation: %s", self._impl)
        log.write("\n")
        return ret

class PvcInitStep(Step):
    """
    A step which initialises partial volume correction
    """
    def run(self, prev_output, log=sys.stdout, fsllog=None, **kwargs):
        """
        Update the MVN from a previous step to include initial estimates
        for PVC parameters
        """
        log.write("Initialising partial volume correction...\n")
        # set the inital GM amd WM values using a simple PV correction
        wm_cbf_ratio = 0.4

        # Modified pvgm map
        temp_pgm = np.copy(self.options["pgm"].data)
        temp_pgm[temp_pgm < 0.2] = 0.2

        # First part of correction psuedo WM CBF term
        prev_ftiss = prev_output["mean_ftiss"].data
        wm_cbf_term = (prev_ftiss * wm_cbf_ratio) * self.options["pwm"].data

        gmcbf_init = (prev_ftiss - wm_cbf_term) / temp_pgm
        wmcbf_init = gmcbf_init * wm_cbf_ratio

        mvn = prev_output["finalMVN"]
        gmcbf_init = Image(gmcbf_init, header=mvn.header)
        wmcbf_init = Image(wmcbf_init, header=mvn.header)

        # HACK: This seems to be required to get the fslpy decorators to write
        # the temporary file correctly
        mask = Image(self.options["mask"].data, header=self.options["mask"].header)

        # load these into the MVN
        mvn = prev_output["finalMVN"]
        from oxasl.wrappers import mvntool
        params = prev_output["paramnames"]
        mvn = mvntool(mvn, params.index("ftiss")+1, output=LOAD, mask=mask, write=True, valim=gmcbf_init, var=0.1, log=fsllog)["output"]
        mvn = mvntool(mvn, params.index("fwm")+1, output=LOAD, mask=mask, write=True, valim=wmcbf_init, var=0.1, log=fsllog)["output"]
        log.write("DONE\n")
        return {"finalMVN" : mvn, "gmcbf_init" : gmcbf_init, "wmcbf_init" : wmcbf_init}



#!/bin/env python
"""
OXASL: ASL processing pipeline
==============================

Overview
--------

This module (and the associated command line tool) provide a 'one stop shop'
for analysing ASL data. They call the various other modules to preprocess,
correct, register, model and calibrate the data, outputting the perfusion
data, and also other estimated parameters.

The processing sequence is approximately:

1. Preprocess ASL data. This involves generation of mean and subtracted images
   and brain extracted versions. Not all of these images will be used later, but
   some might be.

2. Preprocess calibration data. This involves internal motion correction (for
   multi-volume calibration images), and the generation of mean and brain extracted
   versions of the calibration data.

3. Preprocess structural data. This involves generating brain extracted version of
   the structural data, and extracting any existing structural->standard space transformations
   (e.g. if an FSL_ANAT directory is the source of structural information)

4. If requested, the motion correction transformations for the ASL data are determined.
   This is always done so that the middle volume of the
   ASL data is unchanged, although this volume is not always used as the reference.
   If applicable, the calibration->ASL registration is also determined.

5. If requested, the fieldmap distortion correction is determined from the fieldmap
   images

6. All available corrections are applied simultaneously to minimise interpolation

7. If required, a mask is generated from the corrected data

8. Model fitting is carried out to determine ASL space perfusion and other parameter maps

9. The ASL->structural registration is re-done using the perfusion image as the ASL reference.

10. If applicable, calibration is applied to the raw perfusion images

11. Raw and calibrated output images are generated in all output spaces (asl, structural standard)

"""
import sys
import os
import traceback

# Quick-and-dirty plugin imports - to be replaced with entry points and a defined plugin api
# at some point
try:
    import oxasl_ve
except ImportError:
    oxasl_ve = None

try:
    import oxasl_mp
except ImportError:
    oxasl_mp = None

try:
    import oxasl_deblur
except ImportError:
    oxasl_deblur = None

try:
    import oxasl_enable
except ImportError:
    oxasl_enable = None

try:
    import oxasl_multite
except ImportError:
    oxasl_multite = None

from oxasl import *
from oxasl.options import AslOptionParser, GenericOptions, OptionGroup
from oxasl.reporting import LightboxImage

def add_options(parser):
    g = OptionGroup(parser, "General Pipeline Options")
    g.add_option("--wp", help="Analysis which conforms to the 'white papers' (Alsop et al 2014)", action="store_true", default=False)
    g.add_option("--mc", help="Motion correct data", action="store_true", default=False)
    g.add_option("--noreorient", help="Do not reorient ASL/calibration data to standard orientation", action="store_true", default=False)
    if oxasl_enable:
        g.add_option("--use-enable", help="Use ENABLE preprocessing step", action="store_true", default=False)
    if oxasl_deblur:
        g.add_option("--deblur", help="Perform deblurring of ASL and calibration data", action="store_true", default=False)
    parser.add_option_group(g)

    g = OptionGroup(parser, "Model fitting options")
    g.add_option("--fixbat", dest="inferbat", help="Fix bolus arrival time", action="store_false", default=True)
    g.add_option("--batsd", help="Bolus arrival time standard deviation (s) - default 1.0 for multi-PLD, 0.1 otherwise", type=float)
    g.add_option("--fixbolus", "--fixtau", dest="infertau", help="Fix bolus duration", action="store_false")
    g.add_option("--art-off", "--artoff", dest="inferart", help="Do not infer arterial component", action="store_false", default=True)
    g.add_option("--spatial-off", "--spatialoff", dest="spatial", help="Do not include adaptive spatial smoothing on CBF", action="store_false", default=True)
    g.add_option("--infertexch", help="Infer exchange time (multi-TE data only)", action="store_true", default=False)
    g.add_option("--infert1", help="Infer T1 value", action="store_true", default=False)
    g.add_option("--infert2", help="Infer T2 value (multi-TE data only)", action="store_true", default=False)
    g.add_option("--t1im", help="Voxelwise T1 tissue estimates", type="image")
    g.add_option("--batim", "--attim", help="Voxelwise BAT (ATT) estimates in seconds", type="image")
    parser.add_option_group(g)
    
    g = OptionGroup(parser, "Physiological parameters (all have default values from literature)")
    g.add_option("--bat", help="Estimated bolus arrival time (s) - default=0.7 (pASL), 1.3 (cASL)", type=float)
    g.add_option("--t1", "--t1t", help="Tissue T1 (s)", type=float, default=1.3)
    g.add_option("--t2", "--t2t", help="Tissue T2 (ms)", type=float, default=50)
    g.add_option("--t2s", help="Tissue T2* (ms)", type=float, default=20)
    g.add_option("--t1b", help="Blood T1 (s)", type=float, default=1.65)
    g.add_option("--t2b", help="Blood T2 (ms) - Lu et al. 2012 MRM 67:42-49, 3T during normoxia", type=float, default=150)
    g.add_option("--t2sb", help="Blood T2* (ms) - Petersen 2006 MRM 55(2):219-232", type=float, default=50)
    parser.add_option_group(g)

def main():
    """
    Entry point for oxasl command line tool
    """
    debug = True
    wsp = None
    try:
        parser = AslOptionParser(usage="oxasl -i <asl_image> [options]", version=__version__)
        parser.add_category(image.Options())
        parser.add_category(struc.Options())
        add_options(parser)
        m0.add_options(parser)
        parser.add_category(reg.Options())
        distcorr.add_options(parser)
        parser.add_category(senscorr.Options())
        parser.add_category(basil.Options())
        parser.add_category(corrections.Options())
        if oxasl_ve:
            parser.add_category(oxasl_ve.Options())
        if oxasl_mp:
            parser.add_category(oxasl_mp.Options())
        if oxasl_enable:
            parser.add_category(oxasl_enable.Options())
        if oxasl_multite:
            parser.add_category(oxasl_multite.MultiTEOptions())
        if oxasl_deblur:
            parser.add_category(oxasl_deblur.Options())
        parser.add_category(region_analysis.Options())
        parser.add_category(GenericOptions())
        parser.add_category(output.Options())

        options, _ = parser.parse_args()
        debug = options.debug

        # Some oxasl command-line specific defaults
        if (options.calib is not None or options.calib_first_vol) and options.calib_method is None:
            if options.struc is not None:
                options.calib_method = "refregion"
            else:
                options.calib_method = "voxelwise"

        if options.debug:
            options.save_all = True
        options.output_native = True
        options.output_struc = True
        options.save_mask = True

        if options.asldata is None:
            raise RuntimeError("Input ASL file not specified\n")

        if os.path.exists(options.output) and not options.overwrite:
            raise RuntimeError("Output directory exists - use --overwrite to overwrite it")

        wsp = Workspace(savedir=options.output, auto_asldata=True, **vars(options))
        oxasl(wsp)

    except Exception as e:
        sys.stderr.write("ERROR: " + str(e) + "\n")
        if debug:
            traceback.print_exc()
        if wsp is not None:
            wsp.log.write("ERROR: " + str(e) + "\n")
            wsp.log.write(traceback.format_exc() + "\n")
        sys.exit(1)

def oxasl(wsp):
    """
    Main oxasl pipeline script
    """
    wsp.log.write("OXASL version: %s\n" % __version__)
    for plugin in (oxasl_ve, oxasl_mp, oxasl_deblur, oxasl_enable, oxasl_multite):
        if plugin is not None:
            wsp.log.write(" - Found plugin: %s (version %s)\n" % (plugin.__name__, getattr(plugin, "__version__", "unknown")))

    wsp.log.write("\nInput ASL data: %s\n" % wsp.asldata.name)
    wsp.asldata.summary(wsp.log)
    report_asl(wsp)

    # Preprocessing and preparation
    preproc.run(wsp)
    struc.run(wsp)
    moco.run(wsp)
    corrections.run(wsp)
    reg.run(wsp)
    distcorr.run(wsp)
    senscorr.run(wsp)
    corrections.run(wsp)
    mask.run(wsp)

    # Quantification
    filtering.run(wsp)
    prequantify.run(wsp)
    quantify.run(wsp)

    # Calibration and output
    m0.run(wsp)
    rois.run(wsp)
    output.run(wsp)

    # Post processing and reporting
    for quantify_wsp in wsp.quantify_wsps:
        try:
            quantify_name = quantify_wsp[quantify_wsp.index("_"):]
        except ValueError:
            quantify_name = ""
        output_wsp = getattr(wsp, "output%s" % quantify_name)
        region_analysis.run(output_wsp)

    if wsp.save_report:
        reporting.run(wsp)

    _cleanup(wsp)
    wsp.log.write("\nOutput is %s\n" % wsp.savedir)
    wsp.log.write("OXASL - done\n")

def report_asl(wsp):
    """
    Generate a report page about the input ASL data
    """
    page = wsp.report.page("asl")
    page.heading("ASL input data")
    md_table = [(key, value) for key, value in wsp.asldata.metadata_summary().items()]
    page.table(md_table)
    if wsp.pwi is not None:
        img = wsp.pwi
        img_type = "Perfusion-weighted image"
    else:
        img = wsp.asldata.mean()
        img_type = "Mean ASL data"

    page.heading(img_type, level=1)
    page.image("asldata", LightboxImage(img))

def _cleanup(wsp):
    """
    Remove items from the workspace that are not being output. The
    corresponding files will be deleted.
    """
    wsp.log.write("\nDoing cleanup\n")
    if not wsp.save_all:
        if not wsp.save_preproc:
            wsp.log.write(" - Removing preprocessed data\n")
        if not wsp.save_corrected:
            wsp.log.write(" - Removing corrected data\n")
            wsp.corrected = None
        if not wsp.save_corrections:
            wsp.log.write(" - Removing intermediate correction data\n")
            wsp.senscorr = None
            wsp.distcorr = None
            wsp.moco = None
            wsp.topup = None
            wsp.fieldmap = None
        if not wsp.save_reg:
            wsp.log.write(" - Removing registration data\n")
            wsp.reg = None
        if not wsp.save_quantification:
            wsp.log.write(" - Removing model fitting data\n")
            wsp.quantification = None
        if not wsp.save_struc:
            wsp.log.write(" - Removing structural segementation data\n")
            wsp.structural = None
        if not wsp.save_calib:
            wsp.log.write(" - Removing calibration data\n")
            wsp.calibration = None
        if not wsp.save_filter:
            wsp.filter = None
        if not wsp.save_input:
            wsp.input = None
        wsp.rois = None



#!/bin/env python
"""
OXASL - Registration for ASL data

Copyright (c) 2008-2020 Univerisity of Oxford
"""
from __future__ import unicode_literals

import os
import sys
import math
import warnings
import traceback

import numpy as np

from fsl.data.image import Image, defaultExt
import fsl.wrappers as fsl

try:
    import regtricks
except ImportError:
    regtricks = None

from oxasl import __version__, Workspace, struc, brain
from oxasl.options import AslOptionParser, GenericOptions, OptionCategory, OptionGroup, load_matrix
from oxasl.wrappers import epi_reg
from oxasl.reporting import LightboxImage

class Options(OptionCategory):
    """
    OptionCategory which contains options for registration of ASL data to structural image
    """

    def __init__(self, **kwargs):
        OptionCategory.__init__(self, "reg", **kwargs)

    def groups(self, parser):
        groups = []

        group = OptionGroup(parser, "Registration")
        group.add_option("--aslref", "--nativeref", "--regfrom", help="ASL Registration image (e.g. perfusion weighted image)", type="image")
        group.add_option("--aslref-method", "--nativeref-method", "--regfrom-method", help="How to choose the ASL registration reference image - calib=use calibration image, mean=use mean ASL data, pwi=use PWI (mean differenced ASL data)")
        group.add_option("--reg-init-bbr", help="Use BBR for initial as well as final registration - generally used in conjunction with --aslref-method=pwi", action="store_true", default=False)
        group.add_option("--struc-std-fnirt", help="Use FNIRT to do nonlinear structural/std space registration when FSL_ANAT output is not available", action="store_true", default=False)
        #group.add_option("--bbr", dest="do_bbr", help="Include BBR registration step using EPI_REG", action="store_true", default=False)
        #group.add_option("--flirt", dest="do_flirt", help="Include rigid-body registration step using FLIRT", action="store_true", default=True)
        #group.add_option("--flirtsch", help="user-specified FLIRT schedule for registration")
        groups.append(group)

        #group = OptionGroup(parser, "Extra BBR registration refinement")
        #group.add_option("-c", dest="cfile", help="ASL control/calibration image for initial registration - brain extracted")
        #group.add_option("--wm_seg", dest="wm_seg", help="tissue segmenation image for bbr (in structural image space)")
        #groups.append(group)

        #group = OptionGroup(parser, "Distortion correction using fieldmap (see epi_reg)")
        #g.add_option("--nofmapreg", dest="nofmapreg", help="do not perform registration of fmap to T1 (use if fmap already registered)", action="store_true", default=False)
        #groups.append(group)

        #group = OptionGroup(parser, "Deprecated")
        #g.add_option("-r", dest="lowstruc", help="extra low resolution structural image - brain extracted")
        #g.add_option("--inweight", dest="inweight", help="specify weights for input image - same functionality as the flirt -inweight option", type="float")
        #groups.append(group)

        return groups

def run(wsp, redo=False, struc_flirt=True, struc_bbr=False, use_quantification_wsp=None):
    if wsp.reg is None:
        wsp.sub("reg")
        redo = True

    if redo:
        wsp.log.write("\nPerforming registration\n")

        if not struc_bbr and wsp.reg_init_bbr:
            wsp.log.write(" - Using BBR for initial structural registration\n")
            struc_bbr = True

        # Save any previous registration data
        if wsp.reg.aslref is not None:
            idx = 1
            while getattr(wsp.reg, "aslref_old_%i" % idx) is not None:
                idx += 1
            for attr in ("aslref", "asl2struc", "struc2asl"):
                setattr(wsp.reg, "%s_old_%i" % (attr, idx), getattr(wsp.reg, attr))

        get_ref_imgs(wsp, use_quantification_wsp)
        reg_asl2calib(wsp)
        reg_asl2struc(wsp, flirt=struc_flirt, bbr=struc_bbr)
        reg_asl2custom(wsp)

def get_ref_imgs(wsp, use_quantification_wsp=None):
    """
    Get the images that define the various processing 'spaces' and are used for registration
    to/from these spaces. The built in spaces are 'asl' (aka 'native') 'struc' and 'std' (MNI).

    Note that the 'custom' space requires a user-specified reference image and transformation
    from structural space

    aslref defines the 'asl' space

    Optional workspace attributes
    -----------------------------

     - ``aslref`` : User-supplied registration reference image
     - ``aslref_method`` : Method for choosing registration reference image
     - ``asldata`` : Raw ASL data
     - ``calib``   : Calibration image

    Updated workspace attributes
    ----------------------------

     - ``aslref``    : Registration reference image in ASL space
     - ``strucref``     : Registration reference image in structural space
     - ``stdref``       : Registration reference image in standard space
    """
    if wsp.input is not None and wsp.input.aslref_method is None:
        if wsp.asldata is not None and wsp.asldata.iaf in ("tc", "ct"):
            wsp.reg.aslref_method = "mean"
        elif wsp.calib is not None and wsp.asldata is not None and wsp.calib.sameSpace(wsp.asldata):
            wsp.reg.aslref_method = "calib"
        else:
            wsp.reg.aslref_method = "mean"

    if use_quantification_wsp:
        wsp.log.write(" - ASL Registration reference image is PWI image generated by quantification\n")
        pwi = np.copy(use_quantification_wsp.finalstep.mean_ftiss.data)
        pwi[pwi < 0] = 0
        wsp.reg.aslref = Image(pwi, header=use_quantification_wsp.finalstep.mean_ftiss.header)
    elif wsp.input.aslref is not None:
        wsp.log.write(" - ASL Registration reference image supplied by user\n")
        wsp.reg.aslref = wsp.input.aslref
    elif wsp.reg.aslref_method == "mean":
        wsp.log.write(" - ASL Registration reference is mean ASL signal (brain extracted)\n")
        wsp.reg.aslref = brain.brain(wsp, wsp.asldata.mean(), thresh=0.2)
    elif wsp.reg.aslref_method == "calib":
        wsp.log.write(" - ASL Registration reference is calibration image (brain extracted)\n")
        if not wsp.calib.sameSpace(wsp.asldata):
            raise ValueError("Calibration image is not in same space as ASL data - cannot use as registration reference")
        wsp.reg.aslref = brain.brain(wsp, wsp.calib, thresh=0.2)
    elif wsp.reg.aslref_method == "pwi":
        wsp.log.write(" - ASL Registration reference is PWI (brain extracted)\n")
        wsp.reg.aslref = brain.brain(wsp, wsp.asldata.perf_weighted(), thresh=0.2)
    else:
        raise ValueError("Unrecognized aslref_method: %s" % wsp.aslref_method)

    if wsp.input is not None: wsp.reg.calibref = wsp.input.calib
    if wsp.structural is not None: wsp.reg.strucref = wsp.structural.struc
    wsp.reg.stdref = Image(os.path.join(os.environ["FSLDIR"], "data/standard/MNI152_T1_2mm_brain"))

def reg_asl2calib(wsp):
    """
    Register calibration image to ASL space

    Note that this might already have been done as part of motion correction
    """
    if wsp.calib_aslreg:
        wsp.log.write(" - Calibration image already registered to ASL image\n")
    elif wsp.moco is not None and wsp.moco.asl2calib is not None:
        wsp.log.write(" - Calibration image registered to ASL image as part of motion correction\n")
        wsp.reg.asl2calib = wsp.moco.asl2calib
        wsp.reg.calib2asl = wsp.moco.calib2asl
    elif wsp.calib is not None:
        wsp.log.write(" - Registering calibration image to ASL image\n")
        _, wsp.reg.asl2calib = reg_flirt(wsp, wsp.reg.aslref, wsp.calib)
        wsp.reg.calib2asl = np.linalg.inv(wsp.reg.asl2calib)

    if wsp.reg.asl2calib is not None:
        wsp.log.write(" - ASL->Calibration transform\n")
        wsp.log.write(str(wsp.reg.asl2calib) + "\n")
        wsp.log.write(" - Calibration->ASL transform\n")
        wsp.log.write(str(wsp.reg.calib2asl) + "\n")

def reg_asl2custom(wsp):
    """
    Register custom output image to ASL space, via structural. 

    If no output_custom_mat (struc -> custom) has been provided, then
    FLIRT will be used to generate this. The transformation from ASL space
    is the concatenation of asl2struc and struc2custom. 
    """
    if wsp.input.output_custom:
        wsp.reg.customref = wsp.input.output_custom

        if wsp.input.output_custom_mat is not None:
            wsp.reg.struc2custom = np.loadtxt(wsp.input.output_custom_mat)
            wsp.reg.custom2struc = np.linalg.inv(wsp.reg.struc2custom)

        else:  
            wsp.log.write(" - Registering calibration image to ASL image\n")
            _, wsp.reg.custom2struc = reg_flirt(wsp, wsp.reg.customref, wsp.structural.struc)
            wsp.reg.struc2custom = np.linalg.inv(wsp.reg.custom2struc)

        wsp.reg.asl2custom = wsp.reg.struc2custom @ wsp.reg.asl2struc
        wsp.reg.custom2asl = np.linalg.inv(wsp.reg.asl2custom)

def reg_asl2struc(wsp, flirt=True, bbr=False, name="initial"):
    """
    Registration of ASL images to structural image

    :param flirt: If provided, sets whether to use FLIRT registration
    :param bbr: If provided, sets whether to use BBR registration

    Required workspace attributes
    -----------------------------

     - ``aslref``            : Registration reference image in ASL space
     - ``struc``              : Structural image

    Updated workspace attributes
    ----------------------------

     - ``asl2struc``    : ASL->structural transformation matrix
     - ``struc2asl``    : Structural->ASL transformation matrix
     - ``regto``        : ``aslref`` image transformed to structural space
    """
    if wsp.structural is not None and wsp.structural.struc is not None:
        if wsp.struc2asl is not None or wsp.asl2struc is not None:
            wsp.log.write("\nASL->Structural registration provided by user\n")

            wsp.reg.struc2asl = wsp.struc2asl
            wsp.reg.asl2struc = wsp.asl2struc
            if wsp.reg.asl2struc is None:
                wsp.reg.asl2struc = np.linalg.inv(wsp.reg.struc2asl)
            if wsp.reg.struc2asl is None:
                wsp.reg.struc2asl = np.linalg.inv(wsp.reg.asl2struc)
            wsp.reg.regto = change_space(wsp, wsp.reg.aslref, "struc")
        else:
            wsp.log.write("\nRegistering ASL data to structural data\n")
            if flirt:
                wsp.reg.regto, wsp.reg.asl2struc = reg_flirt(wsp, wsp.reg.aslref, wsp.structural.brain, wsp.reg.asl2struc)
            if bbr:
                wsp.reg.regto, wsp.reg.asl2struc = reg_bbr(wsp)

            wsp.reg.struc2asl = np.linalg.inv(wsp.reg.asl2struc)

        wsp.log.write(" - ASL->Structural transform\n")
        wsp.log.write(str(wsp.reg.asl2struc) + "\n")
        wsp.log.write(" - Structural->ASL transform\n")
        wsp.log.write(str(wsp.reg.struc2asl) + "\n")

        name = "final" if bbr else "initial"
        page = wsp.report.page("asl2struc_%s" % name)
        page.heading("%s ASL -> Structural registration" % name.title(), level=0)
        page.heading("Transformation parameters", level=1)
        motion_params = get_transform_params(wsp.reg.asl2struc)
        page.table([
            ["Translation magnitude", "%.3g mm" % motion_params[0]],
            ["Rotation magnitude", "%.3g \N{DEGREE SIGN}" % motion_params[1]],
        ])
        page.heading("ASL->Structural transformation matrix", level=1)
        page.matrix(wsp.reg.asl2struc)
        page.heading("Structural->ASL transformation matrix", level=1)
        page.matrix(wsp.reg.struc2asl)

        if wsp.structural.gm_seg is not None:
            gm_asl = change_space(wsp, wsp.structural.gm_seg, "asl", interp="nn")
            page.heading("GM mask aligned with ASL data", level=1)
            page.image("gm_reg_%s" % name, LightboxImage(gm_asl, bgimage=wsp.reg.aslref))
            wm_asl = change_space(wsp, wsp.structural.wm_seg, "asl", interp="nn")
            page.heading("WM mask aligned with ASL data", level=1)
            page.image("wm_reg_%s" % name, LightboxImage(wm_asl, bgimage=wsp.reg.aslref))

def reg_struc2std(wsp, **kwargs):
    """
    Determine structural -> standard space registration

    Optional workspace attributes
    -----------------------------

     - ``structural.struc``   : Structural image
     - ``fslanat``            : Path to existing FSLANAT data

    Updated workspace attributes
    ----------------------------

     - ``reg.struc2std``    : Structural->MNI transformation matrix - either warp image or FLIRT matrix
     - ``reg.std2struc``    : MNI->structural transformation - either warp image or FLIRT matrix
    """
    if wsp.reg.std2struc is not None:
        return

    if wsp.struc2std_warp is not None:
        wsp.log.write(" - Using user-specified structural->std nonlinear transformation warp\n")
        wsp.reg.struc2std = wsp.struc2std_warp

    elif wsp.struc2std is not None:
        wsp.log.write(" - Using user-specified structural->std linear transformation matrix\n")
        wsp.reg.struc2std = wsp.struc2std
        wsp.log.write(str(wsp.reg.struc2std) + "\n")

    elif wsp.fslanat:
        warp = os.path.join(wsp.fslanat, "T1_to_MNI_nonlin_coeff.nii.gz")
        mat = os.path.join(wsp.fslanat, "T1_to_MNI_lin.mat")
        if os.path.isfile(warp):
            wsp.log.write(" - Using structural->std nonlinear transformation from FSL_ANAT\n")
            wsp.reg.struc2std = Image(warp, loadData=False)
        elif os.path.isfile(mat):
            wsp.log.write(" - Using structural->std linear transformation from FSL_ANAT\n")
            wsp.reg.struc2std = load_matrix(mat)
            wsp.log.write(str(wsp.reg.struc2std) + "\n")

    if wsp.reg.struc2std is None:
        wsp.log.write(" - Registering structural image to standard space using FLIRT\n")
        flirt_result = fsl.flirt(wsp.structural.brain, os.path.join(os.environ["FSLDIR"], "data/standard/MNI152_T1_2mm_brain"), omat=fsl.LOAD)
        wsp.reg.struc2std = flirt_result["omat"]

        if wsp.struc_std_fnirt:
            wsp.log.write(" - Registering structural image to standard space using FNIRT\n")
            fnirt_result = fsl.fnirt(wsp.structural.brain, aff=wsp.reg.struc2std, config="T1_2_MNI152_2mm.cnf", cout=fsl.LOAD)
            wsp.reg.struc2std = fnirt_result["cout"]

    if isinstance(wsp.reg.struc2std, Image):
        # Calculate the inverse warp using INVWARP
        invwarp_result = fsl.invwarp(wsp.reg.struc2std, wsp.structural.struc, out=fsl.LOAD)
        wsp.reg.std2struc = invwarp_result["out"]
    else:
        wsp.reg.std2struc = np.linalg.inv(wsp.reg.struc2std)

def get_img_space(wsp, img):
    """
    Find out what image space an image is in
    
    Note that this only compares the voxel->world transformation matrix to the
    reference image for each space. It is quite possible for two images to be in
    the same space but not be registered to one another. In this case, 
    the returned space may not be accurate when determining whether a registration
    is required.

    :param wsp: Workspace object
    :param img: Image
    :return: Name of image space for ``img``, e.g. ``asl``, ``struc``
    """ 
    img_space = None
    for space in ('asl', 'calib', 'struc', 'std', 'custom'):
        ref = getattr(wsp.reg, "%sref" % space)
        if ref is not None and img.sameSpace(ref):
            img_space = space
            break

    if img_space is None:
        raise RuntimeError("Could not determine space for image: %s" % str(img))
    return img_space

def change_space(wsp, img, target_space, source_space=None, **kwargs):
    """
    Convert an image to a different space
    
    Note that while the source space can be determined from the image, this may
    not be correct if images (e.g. ASL and calibration) share the same voxel->world
    transformation but still need registration to one another

    :param wsp: Workspace object
    :param img: Image
    :param target_space: Either an Image in the target space, or the name of the target space
    :param src_space: If specified, explicit indication of source image space
    """
    # Source and target space can be specified directly or determined from image
    if source_space is None:
        source_space = get_img_space(wsp, img)
    if isinstance(target_space, Image):
        target_space = get_img_space(wsp, target_space)

    # Backwards compatiblity for naming convention
    if source_space == "native":
        source_space = "asl"
    if target_space == "native":
        target_space = "asl"

    target_ref = getattr(wsp.reg, "%sref" % target_space)
    if target_ref is None:
        raise RuntimeError("Couldn't find reference image for target space: %s" % target_space)

    if source_space == target_space:
        # Nothing to be done - image is already in target space
        return img

    if source_space == "std" or target_space == "std":
        # Calculating the nonlinear standard space registration is slow so we only do
        # it if necessary
        reg_struc2std(wsp, **kwargs)

    # For ASL to/from std space, go via structural image using a pre/post matrix
    # Not necessary for custom space because we already have asl2custom and custom2asl
    # from matrix multiplication
    if source_space == "asl" and target_space == "std":
        tform = wsp.reg.struc2std
        kwargs["premat"] = wsp.reg.asl2struc
    elif source_space == "std" and target_space == "asl":
        tform = wsp.reg.std2struc
        kwargs["postmat"] = wsp.reg.struc2asl
    else:
        tform = getattr(wsp.reg, "%s2%s" % (source_space, target_space))

    if tform is None:
        raise RuntimeError("No registration available for transform %s->%s" % (source_space, target_space))

    return transform(wsp, img, tform, target_ref, **kwargs)

def transform(wsp, img, trans, ref, use_flirt=False, interp="trilinear", paddingsize=1, premat=None, postmat=None, mask=False, mask_thresh=0.5):
    """
    Transform an image

    :param wsp: Workspace, used for logging only
    :param img: Image to transform
    :param trans: Transformation matrix or warp image
    :param ref: Target space reference image
    :param use_flirt: Use flirt to apply the transformation which must be a matrix
    :param interp: Interpolation method
    :param paddingsize: Padding size in pixels
    :param premat: If trans is a warp, this can be set to a pre-warp affine transformation matrix

    :return: Transformed Image object
    """
    if trans is None:
        raise ValueError("Transformation matrix not available - has registration been performed?")

    have_warp = isinstance(trans, Image)
    if use_flirt and have_warp:
        raise ValueError("Cannot transform using Flirt when we have a warp")

    if regtricks is not None and have_warp:
        # For now we only use regtricks for nonlinear transformations. It could be used for any but doing so leads to significant
        # numerical differences compared to previous releases. Part of the reason may be some artifacts in the regtricks output
        return transform_regtricks(wsp, img, trans, ref, use_flirt=use_flirt, interp=interp, premat=premat, postmat=postmat, mask=mask, mask_thresh=mask_thresh)

    if use_flirt:
        if interp == "nn":
            interp = "nearestneighbour"
        ret = fsl.applyxfm(img, ref, trans, out=fsl.LOAD, interp=interp, paddingsize=paddingsize, log=wsp.fsllog)["out"]
    else:
        if have_warp:
            kwargs = {"warp" : trans, "premat" : premat, "rel" : True, "postmat" : postmat}
        else:
            kwargs = {"premat" : trans}
            if premat is not None:
                kwargs["premat"] = np.dot(kwargs["premat"], premat)
            elif postmat is not None:
                kwargs["premat"] = np.dot(postmat, kwargs["premat"])
        ret = fsl.applywarp(img, ref, out=fsl.LOAD, interp=interp, paddingsize=paddingsize, super=True, superlevel="a", log=wsp.fsllog, **kwargs)["out"]

    if mask:
        # Binarise mask images
        ret = Image((ret.data > mask_thresh).astype(np.int32), header=ret.header)
    return ret

def transform_regtricks(wsp, img, trans, ref, use_flirt=False, interp="trilinear", premat=None, postmat=None, mask=False, mask_thresh=0.5):
    """
    Transform an image using Regtricks for better performance relative to applywarp

    :param wsp: Workspace, used for logging only
    :param img: Image to transform
    :param trans: Transformation matrix or warp image
    :param ref: Reference image
    :param use_flirt: Use flirt to apply the transformation which must be a matrix
    :param interp: Interpolation method
    :param premat: If trans is a warp, this can be set to a pre-warp affine transformation matrix

    :return: Transformed Image object
    """
    if interp == "nn":
        order = 0
    else:
        order = 1

    have_warp = isinstance(trans, Image)
    if use_flirt and have_warp:
        raise ValueError("Cannot transform using Flirt when we have a warp")
    elif use_flirt:
        rt_trans = regtricks.Registration.from_flirt(trans, src=img.nibImage, ref=ref.nibImage)
    else:
        if have_warp:
            if premat is not None:
                # asl2std
                rt_premat = regtricks.Registration.from_flirt(premat, src=img.nibImage, ref=wsp.reg.strucref.nibImage)
                rt_warp = regtricks.NonLinearRegistration.from_fnirt(trans.nibImage, src=wsp.reg.strucref.nibImage, ref=ref.nibImage)
                rt_trans = regtricks.chain(rt_premat, rt_warp)
            elif postmat is not None:
                # std2asl
                rt_warp = regtricks.NonLinearRegistration.from_fnirt(trans.nibImage, src=img.nibImage, ref=wsp.reg.strucref.nibImage)
                rt_postmat = regtricks.Registration.from_flirt(postmat, src=wsp.reg.strucref.nibImage, ref=ref.nibImage)
                rt_trans = regtricks.chain(rt_warp, rt_postmat)
            else:
                rt_trans = regtricks.NonLinearRegistration.from_fnirt(trans.nibImage, src=img.nibImage, ref=ref.nibImage)
        else:
            if premat is not None:
                # asl2std
                rt_premat = regtricks.Registration.from_flirt(premat, src=img.nibImage, ref=wsp.reg.strucref.nibImage)
                rt_mat = regtricks.Registration.from_flirt(trans, src=wsp.reg.strucref.nibImage, ref=ref.nibImage)
                rt_trans = regtricks.chain(rt_premat, rt_mat)
            elif postmat is not None:
                # std2asl
                rt_mat = regtricks.Registration.from_flirt(trans, src=img.nibImage, ref=wsp.reg.strucref.nibImage)
                rt_postmat = regtricks.Registration.from_flirt(postmat, src=wsp.reg.strucref.nibImage, ref=ref.nibImage)
                rt_trans = regtricks.chain(rt_mat, rt_postmat)
            else:
                if trans.shape[0] != 4:
                    rt_trans = regtricks.MotionCorrection.from_mcflirt(trans, src=img.nibImage, ref=ref.nibImage)
                else:
                    rt_trans = regtricks.Registration.from_flirt(trans, src=img.nibImage, ref=ref.nibImage)

    wsp.log.write(" - Transforming image using regtricks\n")
    ret = Image(rt_trans.apply_to_image(img.nibImage, ref=ref.nibImage), order=order)
    if mask:
        # Binarise mask images
        ret = Image((ret.data > mask_thresh).astype(np.int32), header=ret.header)
    return ret

def reg_flirt(wsp, img, ref, initial_transform=None):
    """
    Register low resolution ASL or calibration data to a high resolution
    structural image using Flirt rigid-body registration

    The brain extracted structural image is used as the reference image. If
    this is not supplied, BET will be run on the whole head structural image.

    :param reg_img: Data to register, e.g. PWI or calibration image. Normally would be brain extracted
    :param struc_brain_img: Brain-extracted structural image

    Optional keyword arguments:

    :param inweight:
    :param init: Initial transform matrix
    :param schedule: FLIRT transform schedule file (default: xyztrans.sch")
    :param dof: FLIRT degrees of freedom

    :return Tuple of registered image, transform matrix
    """
    wsp.log.write(" - Registering image: %s using FLIRT\n" % img.name)

    # Step 1: 3D translation only
    flirt_opts = {
        "schedule" : os.path.join(os.environ["FSLDIR"], "etc", "flirtsch", "xyztrans.sch"),
        "init" : initial_transform,
        "inweight" : wsp.inweight,
        "log" : wsp.fsllog,
    }
    step1_trans = fsl.flirt(img, ref, omat=fsl.LOAD, **flirt_opts)["omat"]

    # Step 2: 6 DOF transformation with small search region
    flirt_opts.update({
        "schedule" : os.path.join(os.environ["FSLDIR"], "etc", "flirtsch", wsp.ifnone("flirtsch", "simple3D.sch")),
        "init" : step1_trans,
        "dof" : wsp.ifnone("dof", 6),
    })
    flirt_result = fsl.flirt(img, ref, out=fsl.LOAD, omat=fsl.LOAD, **flirt_opts)

    return flirt_result["out"], flirt_result["omat"]

def reg_bbr(wsp):
    """
    Perform BBR registration

    :param reg_img: Data to register, e.g. PWI or calibration image. Normally would be brain extracted
    :param struc_img: Structural image
    :param struc_brain_img: Brain-extracted structural image

    Optional keyword arguments:

    :param inweight:
    :param init: Initial transform matrix

    Optional keyword arguments for fieldmap distortion correction:

    :param fmap: Fieldmap image
    :param fmapmag: Fieldmap magnitude image
    :param fmapmagbrain: Fieldmap magnitude image - brain extracted
    :param pedir: Phase encoding direction (x, -x, y, -y, z, -z)
    :param echospacing: Echo spacing

    :return Tuple of registered image, transform matrix
    """
    wsp.log.write(" - BBR registration using epi_reg...")
    # Windows can't run epi_reg as it's a batch script. Use our experimental python
    # implementation but use the standard epi_reg on other platforms until the python
    # version is better tested
    if sys.platform.startswith("win"):
        import oxasl.epi_reg as pyepi
        result = pyepi.epi_reg(wsp, wsp.reg.aslref)
    else:
        result = epi_reg(epi=wsp.reg.aslref, t1=wsp.structural.struc, t1brain=wsp.structural.brain, out=fsl.LOAD, wmseg=wsp.structural.wm_seg, init=wsp.reg.asl2struc, inweight=wsp.inweight, log=wsp.fsllog)
    return result["out%s" % defaultExt()], result["out"]

    #OUTPUT
    #echo "Saving FINAL output"
    #if [ -z $finalonly ]; then
	#cp $outdir/asl2struct.mat $outdir/asl2struct_init.mat # save the initial transformation matrix to allow chekcing if this part failed
    #fi
    #cp $tempdir/low2high_final.mat $outdir/asl2struct.mat #the transformation matrix from epi_reg - this overwrites the version from MAIN registration
    #convert_xfm -omat $outdir/struct2asl.mat -inverse $outdir/asl2struct.mat #often useful to have the inverse transform, so calcuate it
    #if [ ! -z $fmap ]; then
	#imcp $tempdir/low2high_final_warp $outdir/asl2struct_warp #the warp from epi_reg
    #fi
    #imcp $tempdir/low2high_final $outdir/asl2struct # save the transformed image to check on the registration
    #
    # # copy the edge image from epi_reg output as that is good for visualisation
    # imcp $wm_seg $outdir/wm_seg
    #imcp $tempdir/low2high_final_fast_wmedge $outdir/tissedge

def get_transform_params(mat):
    """
    Get motion parameters from a Flirt motion correction matrix

    This is done under the assumption that the matrix may contain
    rotation, translation and possibly minor scaling but no reflection,
    shear etc. So the output could be incorrect for some extreme
    correction matrices, but this probably indicates an error in the
    registration process. We wrap the whole thing in a try block so
    if anything goes horribly wrong it does not at least stop the
    pipeline running

    See http://en.wikipedia.org/wiki/Rotation_matrix for details
    of the rotation calculation.

    :return: Tuple of magnitude of translation, angle and rotation axis
    """
    if tuple(mat.shape) != (4, 4):
        raise ValueError("Not a 4x4 Flirt matrix")

    try:
        # Extract scales - last one is the magnitude of the translation
        scales = np.linalg.norm(mat[:3, :], axis=0)

        # Normalise unit vectors by scaling before extracting rotation
        mat[:, 0] /= scales[0]
        mat[:, 1] /= scales[1]
        mat[:, 2] /= scales[2]

        # Rotation axis
        rot_axis = np.array([
            mat[2, 1] - mat[1, 2],
            mat[0, 2] - mat[2, 0],
            mat[1, 0] - mat[0, 1],
        ], dtype=np.float32)

        # Rotation angle - note that we need to check the sign
        trace = np.trace(mat[:3, :3])
        costheta = (float(trace)-1) / 2
        sintheta = math.sqrt(1-costheta*costheta)
        theta = math.acos(costheta)
        test_element = rot_axis[1]*rot_axis[0]*(1-costheta) + rot_axis[2]*sintheta
        if np.abs(test_element - mat[1, 0]) > np.abs(test_element - mat[0, 1]):
            theta = -theta
        return scales[-1], math.degrees(theta), rot_axis
    except:
        warnings.warn("Error extracting motion parameters from transformation matrix - check registration/moco looks OK!")
        traceback.print_exc()
        return 1, 0, [0, 0, 1]

def main():
    """
    Entry point for command line tool
    """
    try:
        parser = AslOptionParser(usage="asl_reg [options]", version=__version__)
        parser.add_category(Options())
        parser.add_category(struc.StructuralImageOptions())
        parser.add_category(GenericOptions())

        options, _ = parser.parse_args(sys.argv)
        wsp = Workspace(**vars(options))

        if not options.aslref:
            sys.stderr.write("Input file not specified\n")
            parser.print_help()
            sys.exit(1)

        reg_asl2struc(wsp, wsp.do_flirt, wsp.do_bbr)
        if wsp.output:
            wsp.reg.regto.save(wsp.output)
        if wsp.reg.asl2struc:
            with open(wsp.omat, "w") as transform_file:
                for row in wsp.reg.asl2struc:
                    transform_file.write(" ".join(["%f" % val for val in row]) + "\n")

    except ValueError as exc:
        sys.stderr.write("ERROR: " + str(exc) + "\n")
        sys.exit(1)

if __name__ == "__main__":
    main()



#!/bin/env python
"""
OXASL - Distortion correction module

The functions in this module calculate non-linear distortion correction transformations
to apply to the ASL and calibration images.

Currently two sources of transformation exist:

 - Fieldmap-based distortion correction. This generates a nonlinear warp image
   in structural space which is then transformed to ASL space

 - Phase encoding reversed (CBLIP) distortion correction using TOPUP. This generates
   a nonlinear warp image in ASL space

Copyright (c) 2008-2020 Univerisity of Oxford
"""
from __future__ import unicode_literals

import sys

import numpy as np

import fsl.wrappers as fsl
from fsl.data.image import Image

from oxasl import reg
from oxasl.options import OptionGroup
from oxasl.reporting import LightboxImage
from oxasl.wrappers import epi_reg

def add_options(parser):
    g = OptionGroup(parser, "Distortion correction using fieldmap")
    g.add_option("--fmap", help="fieldmap image (in rad/s)", type="image")
    g.add_option("--fmapmag", help="fieldmap magnitude image - wholehead extracted", type="image")
    g.add_option("--fmapmagbrain", help="fieldmap magnitude image - brain extracted", type="image")
    g.add_option("--nofmapreg", help="Do not perform registration of fmap to T1 (use if fmap already in T1-space)", action="store_true", default=False)
    parser.add_option_group(g)
    
    g = OptionGroup(parser, "Distortion correction using phase-encode-reversed calibration image (TOPUP)")
    g.add_option("--cblip", help="phase-encode-reversed (blipped) calibration image", type="image")
    parser.add_option_group(g)
    
    g = OptionGroup(parser, "General distortion correction options")
    g.add_option("--echospacing", help="Effective EPI echo spacing (sometimes called dwell time) - in seconds", type=float)
    g.add_option("--pedir", help="Phase encoding direction, dir = x/y/z/-x/-y/-z")
    g.add_option("--gdc-warp", "--gdcwarp", help="Additional warp image for gradient distortion correction - will be combined with fieldmap or TOPUP distortion correction", type="image")
    parser.add_option_group(g)
        
def run(wsp):
    wsp.sub("distcorr")
    wsp.log.write("\nCalculating distortion corrections\n")
    get_fieldmap_correction(wsp.distcorr)
    get_cblip_correction(wsp.distcorr)
    get_gdc_correction(wsp.distcorr)

def get_cblip_correction(wsp):
    """
    Get the cblip based distortion correction warp

    Required workspace attributes
    -----------------------------

     - ``calib`` : Calibration image
     - ``cblip`` : Phase-encode-reversed calibration image
     - ``echospacing`` :
     - ``pedir`` :

    Updated workspace attributes
    ----------------------------

     - ``cblip_warp``    : CBLIP Distortion correction warp image
    """
    if wsp.topup is not None:
        return
    elif wsp.cblip is None:
        wsp.log.write(" - No CBLIP images provided for distortion correction\n")
        return

    wsp.sub("topup")
    wsp.log.write(" - Calculating distortion Correction using TOPUP\n")

    topup_params = {
        "x"  : [[1, 0, 0, -99], [-1, 0, 0, -99]],
        "-x" : [[-1, 0, 0, -99], [1, 0, 0, -99]],
        "y"  : [[0, 1, 0, -99], [0, -1, 0, -99]],
        "-y" : [[0, -1, 0, -99], [0, 1, 0, -99]],
        "z"  : [[0, 0, 1, -99], [0, 0, -1, -99]],
        "-z" : [[0, 0, -1, -99], [0, 0, 1, -99]],
    }
    dim_idx = {
        "x"  : 0, "-x" : 0,
        "y"  : 1, "-y" : 1,
        "z"  : 2, "-z" : 2,
    }
    my_topup_params = np.array(topup_params[wsp.pedir], dtype=np.float32)
    dimsize = wsp.asldata.shape[dim_idx[wsp.pedir]]
    my_topup_params[:, 3] = wsp.echospacing * (dimsize - 1)
    wsp.topup.params = my_topup_params

    # Run TOPUP to calculate correction
    wsp.topup.calib_blipped = reg.change_space(wsp, Image(np.stack((wsp.calib.data, wsp.cblip.data), axis=-1), header=wsp.calib.header), 'asl')
    topup_result = fsl.topup(imain=wsp.topup.calib_blipped, datain=wsp.topup.params, config="b02b0.cnf", out=fsl.LOAD, iout=fsl.LOAD, fout=fsl.LOAD, log=wsp.fsllog)
    wsp.topup.fieldcoef, wsp.topup.movpar = topup_result["out_fieldcoef"], topup_result["out_movpar"]
    wsp.topup.iout = topup_result["iout"]
    wsp.topup.fout = topup_result["fout"]

    page = wsp.report.page("topup")
    page.heading("TOPUP distortion correction", level=0)
    page.text("PE direction: %s" % wsp.pedir)
    page.text("Echo spacing: %f s" % wsp.echospacing)
    page.heading("Correction image", level=1)
    for dim in range(3):
        img = Image(wsp.topup.fieldcoef.data[..., dim], header=wsp.topup.fieldcoef.header)
        page.text("Dimension %i" % dim)
        page.image("fmap_warp%i" % dim, LightboxImage(img))

def get_fieldmap_correction(wsp):
    """
    Get the fieldmap based distortion correction warp

    Required workspace attributes
    -----------------------------

     - ``pwi``          : Perfusion weighted image (generated by preproc_asl)
     - ``fmap``         : Fieldmap image
     - ``fmapmag``      : Fieldmap magnitude image
     - ``fmapmagbrain`` : Fieldmap magnitude brain image
     - ``echospacing``  :
     - ``pedir``        :

    Optional workspace attributes
    -----------------------------

     - ``nofmapreg``       : If True assume fieldmap in structural space

    Updated workspace attributes
    ----------------------------

     - ``fmap_warp``    : Fieldmap distortion correction warp image in ASL space
    """
    if wsp.fmap is None or wsp.fmapmag is None or wsp.fmapmagbrain is None:
        wsp.log.write(" - No fieldmap images for distortion correction\n")
        return
    elif wsp.pedir is None or wsp.echospacing is None:
        wsp.log.write(" -WARNING: Fieldmap images supplied but pedir and echospacing required for distortion correction\n")
        return

    wsp.sub("fieldmap")
    wsp.log.write(" - Calculating distortion correction from fieldmap images using EPI_REG\n")

    epi_reg_opts = {
        "inweight" : wsp.inweight,
        "init" : wsp.reg.asl2struc,
        "fmap" : wsp.fmap,
        "fmapmag" : wsp.fmapmag,
        "fmapmagbrain" : wsp.fmapmagbrain,
        "pedir" : wsp.pedir,
        "echospacing" : "%.10f" % float(wsp.echospacing),
        "nofmapreg" : wsp.ifnone("nofmapreg", False),
    }

    # Windows can't run epi_reg as it's a batch script. Use our experimental python
    # implementation but use the standard epi_reg on other platforms until the python
    # version is better tested
    if sys.platform.startswith("win"):
        import oxasl.epi_reg as pyepi
        result = pyepi.epi_reg(wsp, epi=wsp.reg.aslref, **epi_reg_opts)
    else:
        result = epi_reg(epi=wsp.asldata.perf_weighted(), t1=wsp.structural.struc, t1brain=wsp.structural.brain, out=fsl.LOAD, wmseg=wsp.structural.wm_seg, log=wsp.fsllog, **epi_reg_opts)

    # Occasionally we end up with NaN in the output of epi_reg and this will ruin the entire distcorr warp.
    # So remove any NaN values and replace with zero.
    warp_struc = result["out_warp"]
    wsp.fieldmap.warp_struc = Image(np.nan_to_num(warp_struc.data, nan=0), header=warp_struc.header)
    wsp.fieldmap.asl2struc = result["out"]
    wsp.fieldmap.struc2asl = np.linalg.inv(wsp.fieldmap.asl2struc)

    result = fsl.convertwarp(out=fsl.LOAD, ref=wsp.reg.aslref, warp1=wsp.fieldmap.warp_struc, postmat=wsp.fieldmap.struc2asl, rel=True, log=wsp.fsllog)
    wsp.fieldmap.warp = result["out"]

    page = wsp.report.page("fmap")
    page.heading("Fieldmap distortion correction", level=0)
    page.text("PE direction: %s" % wsp.pedir)
    page.text("Echo spacing: %f s" % wsp.echospacing)
    page.heading("Correction warps", level=1)
    for dim in range(3):
        img = Image(wsp.fieldmap.warp.data[..., dim], header=wsp.fieldmap.warp.header)
        page.text("Dimension %i" % dim)
        page.image("fmap_warp%i" % dim, LightboxImage(img))

def get_gdc_correction(wsp):
    """
    User-specified gradient distortion correction warp assumed to be in ASL space
    """
    wsp.gdc_warp = wsp.gdc_warp



"""
OXASL - Quantification using Fabber

Includes partial volume correction

Copyright (c) 2008-2020 Univerisity of Oxford
"""

import numpy as np

from fsl.data.image import Image

from oxasl import mask, reg
from . import multistep_fit


def run(wsp):
    # Basic non-PVC run
    multistep_fit.run(wsp.sub("basil"))
    wsp.quantify_wsps.append("basil")

    # Re-do registration using PWI as reference
    reg.run(wsp, redo=True, struc_bbr=True, struc_flirt=False, use_quantification_wsp=wsp.basil)

    # Quantification in alternate spaces
    for quantify_space in ("struc", "std", "custom"):
        if wsp.ifnone("quantify_%s" % quantify_space, False):
            quantify_name = "basil_%s" % quantify_space
            quantify_wsp = wsp.sub(quantify_name)
            quantify_wsp.image_space = quantify_space
            multistep_fit.run(quantify_wsp) 
            wsp.quantify_wsps.append(quantify_name)

    # If the user has provided manual PV maps (pvgm and pvgm) then do PVEc, even if they
    # have not explicitly given the --pvcorr option 
    wsp.user_pv_flag = ((wsp.pvwm is not None) and (wsp.pvgm is not None))
    if wsp.pvcorr or wsp.user_pv_flag:
        # Partial volume correction is very sensitive to the mask, so recreate it
        # if it came from the structural image as this requires accurate ASL->Struc registration
        if wsp.rois.mask_src == "struc":
            wsp.rois.mask_orig = wsp.rois.mask
            wsp.rois.mask = None
            mask.run(wsp)

        if wsp.pvcorr or wsp.user_pv_flag:
            _default_pvcorr(wsp)

def _default_pvcorr(wsp):
    # Do partial volume correction fitting
    #
    # FIXME: We could at this point re-apply all corrections derived from structural space?
    # But would need to make sure corrections module re-transforms things like sensitivity map
    
    # Prepare GM and WM partial volume maps from FAST segmentation
    if wsp.user_pv_flag:
        wsp.log.write("\nUsing user-supplied PV estimates\n")
        wsp.structural.wm_pv_asl = wsp.pvwm
        wsp.structural.gm_pv_asl = wsp.pvgm
    elif wsp.structural.struc is not None:
        wsp.structural.wm_pv_asl = reg.change_space(wsp, wsp.structural.wm_pv, "asl")
        wsp.structural.gm_pv_asl = reg.change_space(wsp, wsp.structural.gm_pv, "asl")
    else:
        raise RuntimeError("Can't do partial volume correction without either user PV estimates or structural image")

    wsp.basil_options = wsp.ifnone("basil_options", {})
    wsp.basil_options.update({"pwm" : wsp.structural.wm_pv_asl, 
                                "pgm" : wsp.structural.gm_pv_asl})
    multistep_fit.run(wsp.sub("basil_pvcorr"), prefit=False)
    wsp.quantify_wsps.append("basil_pvcorr")




#!/bin/env python
"""
OXASL - Module to calibrate a perfusion output using previously calculated M0 value or image

Copyright (c) 2008-2020 Univerisity of Oxford
"""
import numpy as np

from fsl.data.image import Image

from oxasl.options import OptionCategory

class Options(OptionCategory):
    """
    Options for calibration
    """

    def __init__(self):
        OptionCategory.__init__(self, "calibrate")

    def groups(self, parser):
        groups = []
        return groups

def run(wsp, perf_img, multiplier=1.0, var=False):
    """
    Do calibration of a perfusion image from a calibration (M0) image or value

    :param wsp: Workspace object
    :param perf_img: Image containing perfusion data to calibrate
    :param multiplier: Scalar multiple to convert output to physical units
    :param alpha: Inversion efficiency
    :param var: If True, assume data represents variance rather than value

    :return: Image containing calibrated data

    Required workspace attributes
    -----------------------------

     - ``m0`` : M0 single value or voxelwise Image
    """
    if not perf_img:
        raise ValueError("Perfusion data cannot be None")
    if not wsp.m0:
        raise ValueError("No calibration data supplied")

    wsp.log.write("\nCalibrating perfusion data: %s\n" % perf_img.name)
    alpha = wsp.ifnone("calib_alpha", 1.0 if wsp.asldata.iaf in ("ve", "vediff") else 0.85 if wsp.asldata.casl else 0.98)
    m0 = wsp.m0
    if isinstance(m0, Image):
        m0 = m0.data

    if var:
        wsp.log.write(" - Treating data as variance - squaring M0 correction, multiplier and inversion efficiency\n")
        m0 = m0**2
        multiplier = multiplier**2
        alpha = alpha**2

    if isinstance(m0, np.ndarray):
        # If M0 is <= zero, make calibrated data zero
        calibrated = np.zeros(perf_img.shape)
        calibrated[m0 > 0] = perf_img.data[m0 > 0] / m0[m0 > 0]
    else:
        calibrated = perf_img.data / m0

    if alpha != 1.0:
        wsp.log.write(" - Using inversion efficiency correction: %f\n" % alpha)
        calibrated /= alpha

    if multiplier != 1.0:
        wsp.log.write(" - Using multiplier for physical units: %f\n" % multiplier)
        calibrated *= multiplier

    perf_calib = Image(calibrated, name=perf_img.name, header=perf_img.header)
    return perf_calib



"""
OXASL - Module to generate a suitable mask for ASL data

Copyright (c) 2008-2020 Univerisity of Oxford
"""
import numpy as np
import scipy as sp

from fsl.data.image import Image

from oxasl import reg
from oxasl.reporting import LightboxImage

def generate_mask(wsp):
    """
    For compatibility
    """
    run(wsp)

def run(wsp):
    """
    Generate mask for ASL data

    - If a ready-made mask image is provided or has already been generated, this is returned
    - If a structural image is provided this will be used. Brain extraction and registration
      will be performed if required
    - If a calibration image is provided, this is used. It is assumed to be in the same space
      as the ASL data
    - If none of the above are present, the ASL data itself is averaged and brain extracted
      to produce the mask

    Required workspace attributes
    -----------------------------

    Formally there are no required attributes, however at least one image must be provided
    which enables a mask to be generated.

    Optional workspace attributes
    -----------------------------

     - ``asldata`` : ASL data image
     - ``mask``    : Existing brain mask
     - ``struc``   : Structural image (wholehead)
     - ``struc_brain``: Already brain-extracted structural image
     - ``asl2struc`` : Existring ASL->Structural space transformation matrix
     - ``calib``   : Calibration image
     - ``aslref`` : ASL registration source image
    """
    if wsp.rois is not None and wsp.rois.mask is not None:
        return

    wsp.sub("rois")
    wsp.log.write("\nGenerating ASL data mask\n")

    # Reporting
    page = wsp.report.page("mask")
    page.heading("Mask generation", level=0)

    if wsp.mask is not None:
        if reg.get_img_space(wsp, wsp.mask) != "asl":
            raise ValueError("Mask supplied is not in ASL space: %s" % wsp.mask.name)
        wsp.rois.mask_src = "user"
        mask_source = "ASL space mask provided by user: %s" % wsp.mask.name
        wsp.rois.mask = wsp.mask
    elif wsp.structural is not None and wsp.structural.struc is not None:
        # Preferred option is to use brain extracted structural
        wsp.rois.mask_src = "struc"
        page.heading("Brain extracted structural image", level=1)
        page.image("struc_brain", LightboxImage(wsp.structural.brain, bgimage=wsp.structural.struc))
        wsp.rois.mask_struc = wsp.structural.brain_mask
        wsp.rois.mask_asl = reg.change_space(wsp, wsp.structural.brain_mask, "asl")
        wsp.rois.mask = Image(sp.ndimage.morphology.binary_fill_holes((wsp.rois.mask_asl.data > 0.25)).astype(np.int32), header=wsp.rois.mask_asl.header)
        mask_source = "generated from brain extracting structural image and registering to ASL space"
    else:
        # Alternatively, use registration image (which will be BETed calibration or mean ASL image)
        wsp.rois.mask_src = "aslref"
        wsp.rois.mask = Image((wsp.reg.aslref.data != 0).astype(np.int32), header=wsp.reg.aslref.header)
        mask_source = "generated from brain extracted registration ASL image"

    wsp.log.write(" - Mask %s\n" % mask_source)

    page.heading("Masked ASL brain image", level=1)
    page.text("Mask was %s" % mask_source)
    page.text("PW ASL image masked by ASL-space mask")

    if wsp.asldata.iaf in ("diff", "tc", "ct"):
        page.image("mask_outline", LightboxImage(wsp.rois.mask, bgimage=wsp.asldata.perf_weighted(), outline=True))
    else:
        page.image("mask_outline", LightboxImage(wsp.rois.mask, bgimage=wsp.asldata.mean(), outline=True))



#!/bin/env python
"""
OXASL - Module to apply Moco/Distortion/sensitivity corrections

This module generates corrected ASL/calibration data from previously calculated
corrections with the minimum of interpolation

Currently the following sources of transformation exist:

 - Motion correction of the ASL data. This generates a series of linear (rigid body)
   transformations in ASL space, one for each ASL volume. If calibration data is also
   present a calibration->ASL transform is also generated as part of this process

 - Fieldmap-based distortion correction. This generates a nonlinear warp image
   in structural space which is then transformed to ASL space

 - Phase encoding reversed (CBLIP) distortion correction using TOPUP. This generates
   a nonlinear warp image in ASL space FIXME calibration space?

 - User-supplied nonlinear warp image for gradient distortion corection

 - Sensitivity correction

Except for the TOPUP correction, all of the above can be combined in a single
transformation to minimise interpolation of the ASL data

Copyright (c) 2008-2020 Univerisity of Oxford
"""
from __future__ import unicode_literals

import tempfile
import shutil

import numpy as np

import fsl.wrappers as fsl
from fsl.data.image import Image

from oxasl import reg
from oxasl.options import OptionCategory
from oxasl.wrappers import fnirtfileutils

class Options(OptionCategory):
    """
    Options for corrections of the input data
    """

    def __init__(self):
        OptionCategory.__init__(self, "corrections")

    def groups(self, parser):
        ret = []
        return ret

def run(wsp):
    """
    Apply distortion and motion corrections to ASL and calibration data

    Required workspace attributes
    -----------------------------

     - ``asldata_orig`` : Uncorrected ASL data image

    Optional workspace attributes
    -----------------------------

     - ``calib_orig``      : Calibration image
     - ``cref_orig``       : Calibration reference image
     - ``cblip_orig``      : Calibration BLIP image
     - ``asldata_mc_mats`` : ASL motion correction matrices
     - ``calib2asl``       : Calibration -> ASL transformation matrix
     - ``distcorr_warp``   : Distortion correction warp image
     - ``gdc_warp``        : Gradient distortion correction warp image

    Updated workspace attributes
    ----------------------------

     - ``asldata``    : Corrected ASL data
     - ``calib``      : Corrected calibration image
     - ``cref``       : Corrected calibration reference image
     - ``cblip``      : Corrected calibration BLIP image
    """
    wsp.sub("corrected")
    wsp.log.write("\nApplying data corrections\n")

    warps, moco_mats = [], None
    
    if wsp.moco is not None:
        wsp.log.write(" - Using motion correction\n")
        moco_mats = wsp.moco.mc_mats

    if wsp.distcorr is not None:
        if wsp.distcorr.fieldmap is not None:
            wsp.log.write(" - Using fieldmap distortion correction\n")
            warps.append(wsp.distcorr.fieldmap.warp)

        if wsp.distcorr.gdc_warp:
            wsp.log.write(" - Using user-supplied GDC warp\n")
            warps.distcorr.append(wsp.gdc_warp)

    if warps:
        kwargs = {}
        for idx, warp in enumerate(warps):
            kwargs["warp%i" % (idx+1)] = warp

        wsp.log.write(" - Converting all warps to single transform and extracting Jacobian\n")
        result = fsl.convertwarp(ref=wsp.reg.aslref, out=fsl.LOAD, rel=True, jacobian=fsl.LOAD, log=wsp.fsllog, **kwargs)
        wsp.corrected.total_warp = result["out"]

        # Calculation of the jacobian for the warp - method suggested in:
        # https://www.jiscmail.ac.uk/cgi-bin/webadmin?A2=FSL;d3fee1e5.0908
        wsp.corrected.warp_coef = fnirtfileutils(wsp.corrected.total_warp, outformat="spline", out=fsl.LOAD, log=wsp.fsllog)["out"]
        jacobian = fnirtfileutils(wsp.corrected.warp_coef, jac=fsl.LOAD, log=wsp.fsllog)["jac"]
        wsp.corrected.jacobian = Image(jacobian.data, header=wsp.corrected.total_warp.header)

    if not warps and moco_mats is None and wsp.senscorr is None:
        wsp.log.write(" - No corrections to apply to ASL data\n")
        wsp.corrected.asldata = wsp.preproc.asldata
    else:
        # Apply all corrections to ASL data - note that we make sure the output keeps all the ASL metadata
        wsp.log.write(" - Applying corrections to ASL data\n")
        asldata_corr = correct_img(wsp, wsp.preproc.asldata, moco_mats)
        wsp.corrected.asldata = wsp.preproc.asldata.derived(asldata_corr.data)

    if wsp.preproc.calib is not None:
        # Apply corrections to calibration images if we have calib2asl registration or any other correction
        if not warps and (wsp.reg is None or wsp.reg.calib2asl is None) and wsp.senscorr is None:
            wsp.log.write(" - No corrections to apply to calibration data\n")
            wsp.corrected.calib = wsp.preproc.calib
            if wsp.cref is not None:
                wsp.corrected.cref = wsp.preproc.cref
            if wsp.cblip is not None:
                wsp.corrected.cblip = wsp.preproc.cblip
        else:
            wsp.log.write(" - Applying corrections to calibration data\n")
            wsp.corrected.calib = correct_img(wsp, wsp.preproc.calib, wsp.reg.calib2asl)

            if wsp.cref is not None:
                wsp.corrected.cref = correct_img(wsp, wsp.preproc.cref, wsp.reg.calib2asl)
            if wsp.cblip is not None:
                wsp.corrected.cblip = correct_img(wsp, wsp.preproc.cblip, wsp.reg.calib2asl)

    if wsp.distcorr is not None and wsp.distcorr.topup is not None:
        wsp.log.write(" - Adding TOPUP distortion correction\n")
        # This can't currently be done using the FSL wrappers - we need the TOPUP output as two prefixed files
        # Only workaround currently is to create a temp directory to store appropriately named input files
        topup_input = tempfile.mkdtemp(prefix="topup_input")
        try:
            wsp.distcorr.topup.fieldcoef.save("%s/topup_fieldcoef" % topup_input)
            movpar_file = open("%s/topup_movpar.txt" % topup_input, "w")
            for row in wsp.distcorr.topup.movpar:
                movpar_file.write("\t".join([str(val) for val in row]) + "\n")
            movpar_file.close()
            # TOPUP does not do the jacobian magntiude correction - so only okay if using voxelwise calibration
            wsp.corrected.calib = fsl.applytopup(wsp.corrected.calib, datain=wsp.distcorr.topup.params, index=1, topup="%s/topup" % topup_input, out=fsl.LOAD, method="jac", log=wsp.fsllog)["out"]
            if wsp.cref is not None:
                wsp.corrected.cref = fsl.applytopup(wsp.corrected.cref, datain=wsp.distcorr.topup.params, index=1, topup="%s/topup" % topup_input, out=fsl.LOAD, method="jac", log=wsp.fsllog)["out"]
            if wsp.cblip is not None:
                wsp.corrected.cblip = fsl.applytopup(wsp.corrected.cblip, datain=wsp.distcorr.topup.params, index=2, topup="%s/topup" % topup_input, out=fsl.LOAD, method="jac", log=wsp.fsllog)["out"]
            post_topup = fsl.applytopup(wsp.corrected.asldata, datain=wsp.distcorr.topup.params, index=1, topup="%s/topup" % topup_input, out=fsl.LOAD, method="jac", log=wsp.fsllog)["out"]
            wsp.corrected.asldata = wsp.corrected.asldata.derived(post_topup.data)
            # FIXME warning below
            # FIXME do we need to correct anything else in ASL or calibration space, e.g. mask, reference region mask
            #if wsp.calib_method != "voxel":
            #    wsp.log.write("WARNING: Using TOPUP does not correct for magntiude using the jocbian in distortion correction")
            #    wsp.log.write("         This is not optimal when not using voxelwise calibration\n")
            #    wsp.log.write("         To avoid this supply structural image(s)\n")
        finally:
            shutil.rmtree(topup_input)

    try:
        wsp.corrected.pwi = wsp.corrected.asldata.perf_weighted()
    except:
        # Ignore - not all data can generate a PWI
        pass

def correct_img(wsp, img, linear_mat):
    """
    Apply combined warp/linear transformations to an image in ASL space

    :param img: fsl.data.image.Image to correct
    :param linear_mat: img->ASL space linear transformation matrix.
    :return: Corrected Image

    If a jacobian is present, also corrects for quantitative signal magnitude as volume has been locally scaled

    FIXME there are slight differences to oxford_asl here due to use of spline interpolation rather than
    applyxfm4D which uses sinc interpolation.

    Required workspace attributesd
    -----------------------------

     - ``asldata_mean`` : Mean ASL image used as reference space

    Optional workspace attributes
    -----------------------------

     - ``total_warp``      : Combined warp image
     - ``jacobian``        : Jacobian associated with warp image
     - ``senscorr``        : Sensitivity correction
    """
    if wsp.corrected.total_warp is not None:
        img = reg.transform(wsp, img, trans=wsp.corrected.total_warp, ref=wsp.preproc.aslspace, premat=linear_mat)
    elif linear_mat is not None:
        img = reg.transform(wsp, img, trans=linear_mat, ref=wsp.preproc.aslspace)

    if wsp.corrected.jacobian is not None:
        wsp.log.write(" - Correcting for local volume scaling using Jacobian\n")
        jdata = wsp.corrected.jacobian.data
        if img.data.ndim == 4:
            # Required to make broadcasting work
            jdata = jdata[..., np.newaxis]
        img = Image(img.data * jdata, header=img.header)

    if wsp.senscorr is not None:
        wsp.log.write(" - Applying sensitivity correction\n")
        sens_data = reg.change_space(wsp, wsp.senscorr.sensitivity, img).data
        if img.ndim == 4:
            sens_data = sens_data[..., np.newaxis]
        img = Image(img.data / sens_data, header=img.header)

    return img



"""
oxasl.gui.run_box.py

Component of the OXASL GUI which runs the OXASL pipeline

Copyright (c) 2019 University of Oxford
"""
import sys
import traceback
from threading import Thread

import nibabel as nib

import wx
from wx.lib.pubsub import pub

from oxasl import Workspace
from oxasl.pipeline import oxasl
from oxasl.gui.widgets import OptionError

class LogWriter():
    """
    File-like interface to send OXASL log output to the GUI output window
    """
    def write(self, line):
        wx.CallAfter(pub.sendMessage, "run_stdout", line=line)

    def flush(self):
        pass

class OxaslRunner(Thread):
    """
    Runs OXASL pipeline in a background thread
    """
    def __init__(self, options):
        Thread.__init__(self)
        self.options = options

    def run(self):
        ret = -1
        try:
            wsp = Workspace(log=LogWriter(), **self.options)
            oxasl(wsp)
            ret = 0
        finally:
            wx.CallAfter(pub.sendMessage, "run_finished", retcode=ret)

class AslRun(wx.Frame):
    """
    Runs the OXASL pipeline, displaying output in a window
    """

    def __init__(self, parent, run_btn, run_label):
        wx.Frame.__init__(self, parent, title="Run", size=(600, 400), style=wx.DEFAULT_FRAME_STYLE)

        self.options = None
        self.run_btn = run_btn
        self.run_btn.Bind(wx.EVT_BUTTON, self.dorun)
        self.run_label = run_label
        self.preview_data = None

        self.sizer = wx.BoxSizer(wx.VERTICAL)
        self.output_text = wx.TextCtrl(self, style=wx.TE_READONLY | wx.TE_MULTILINE)
        font = wx.Font(8, wx.TELETYPE, wx.NORMAL, wx.NORMAL)
        self.output_text.SetFont(font)
        self.sizer.Add(self.output_text, 1, flag=wx.EXPAND)

        self.SetSizer(self.sizer)
        self.Bind(wx.EVT_CLOSE, self.close)
        pub.subscribe(self.write_output, "run_stdout")
        pub.subscribe(self.finished, "run_finished")

    def write_output(self, line):
        """
        Write a line onto the log output window
        """
        self.output_text.AppendText(line)

    def close(self, _):
        """
        Hide the log output window
        """
        self.Hide()

    def finished(self, retcode):
        """
        Called by background thread when pipeline has finished (successful or not)
        """
        if retcode != 0:
            self.write_output("\nWARNING: command failed\n")
        self.update()

    def dorun(self, _):
        """
        When 'Run' button clicked, run the pipeline displaying the output in the
        log window
        """
        if self.options:
            self.Show()
            self.Raise()
            self.output_text.Clear()
            self.run_btn.Enable(False)
            self.run_label.SetForegroundColour(wx.Colour(0, 0, 128))
            self.run_label.SetLabel("Running - Please Wait")
            OxaslRunner(self.options).start()

    def update(self):
        """
        Get the OXASL options, and check they are valid. If not valid,
        display the first error in the status label
        """
        self.options = None
        try:
            self.options = self.get_options()
            self.run_label.SetForegroundColour(wx.Colour(0, 128, 0))
            self.run_label.SetLabel("Ready to Go")
            self.run_btn.Enable(True)
        except (OptionError, ValueError, nib.filebasedimages.ImageFileError) as exc:
            self.run_btn.Enable(False)
            self.run_label.SetForegroundColour(wx.Colour(255, 0, 0))
            self.run_label.SetLabel(str(exc))
        except:
            # Any other exception is a program bug - report it to STDERR
            self.run_btn.Enable(False)
            self.run_label.SetForegroundColour(wx.Colour(255, 0, 0))
            self.run_label.SetLabel("Unexpected error - see console and report as a bug")
            traceback.print_exc(sys.exc_info()[1])

    def get_preview_data(self):
        """
        Get a perfusion weighted image for the preview
        """
        asldata = self.input.options()["asldata"]
        return asldata.perf_weighted().data

    def get_options(self):
        """
        Get OXASL options

        Exception text is reported by the GUI
        """
        options = {
            "output_native" : True,
            "output_struc" : True,
            "save_mask" : True,
        }
        options.update(self.input.options())
        options.update(self.analysis.options())
        options.update(self.structure.options())
        options.update(self.distcorr.options())
        options.update(self.calibration.options())
        #print(options)

        return options



"""
OXASL - Module to generate additional ROIs for analysis

Copyright (c) 2008-2020 Univerisity of Oxford
"""
import numpy as np

from fsl.data.image import Image
from fsl.data.atlases import AtlasRegistry

from oxasl import reg

def run(wsp):
    """
    Generate additional ROIs used for reporting and region analysis

    wsp.rois.gm_asl: GM mask in ASL space (>0.5 probability)
    wsp.rois.wm_asl: WM mask in ASL space (>0.5 probability)
    wsp.rois.pure_gm_asl: 'pure' GM mask in ASL space (default > 0.8 probability, but uses pure_gm_thresh option)
    wsp.rois.pure_wm_asl: 'pure' WM mask in ASL space (default > 0.9 probability, but uses pure_gm_thresh option)
    wsp.rois.cortial_gm_asl: Cortical GM mask in ASL space, same as pure_gm_asl but masked to remove subcortical structures
    wsp.rois.cerebral_wm_asl: Cererbral WM mask in ASL space, same as pure_wm_asl but masked to remove subcortical structures
    """
    if wsp.structural.struc is not None:
        # Get the cortex 
        atlases = AtlasRegistry()
        atlases.rescanAtlases()
        atlas = atlases.loadAtlas("harvardoxford-subcortical", loadSummary=False, resolution=2)
        wsp.rois.cortex_std = Image(np.mean(atlas.data[..., 0:2] + atlas.data[..., 11:13], axis=-1) * 2, header=atlas.header)
        cortex_asl = reg.change_space(wsp, wsp.rois.cortex_std, "asl")
        wsp.rois.cortex_asl = Image((cortex_asl.data > 50).astype(int), header=cortex_asl.header)

        if wsp.structural.gm_pv_asl is None:
            wsp.structural.gm_pv_asl = reg.change_space(wsp, wsp.structural.gm_pv, "asl")
            wsp.structural.wm_pv_asl = reg.change_space(wsp, wsp.structural.wm_pv, "asl")
            wsp.structural.csf_pv_asl = reg.change_space(wsp, wsp.structural.csf_pv, "asl")

        gm = np.asarray(wsp.structural.gm_pv_asl.data)
        wm = np.asarray(wsp.structural.wm_pv_asl.data)

        wsp.rois.pure_gm_thresh = wsp.ifnone("pure_gm_thresh", 0.8)
        wsp.rois.pure_wm_thresh = wsp.ifnone("pure_wm_thresh", 0.9)
        some_gm, some_wm = gm > 0.5, wm > 0.5
        pure_gm, pure_wm = gm > wsp.rois.pure_gm_thresh, wm > wsp.rois.pure_wm_thresh
        wsp.rois.gm_asl = Image(some_gm.astype(int), header=wsp.structural.gm_pv_asl.header)
        wsp.rois.pure_gm_asl = Image(pure_gm.astype(int), header=wsp.structural.gm_pv_asl.header)
        wsp.rois.wm_asl = Image(some_wm.astype(int), header=wsp.structural.wm_pv_asl.header)
        wsp.rois.pure_wm_asl = Image(pure_wm.astype(int), header=wsp.structural.wm_pv_asl.header)
        wsp.rois.cortical_gm_asl = Image(np.logical_and(pure_gm, cortex_asl.data).astype(int), header=wsp.structural.gm_pv_asl.header)
        wsp.rois.cerebral_wm_asl = Image(np.logical_and(pure_wm, cortex_asl.data).astype(int), header=wsp.structural.wm_pv_asl.header)



"""
oxasl.gui.calibration_tab.py

Copyright (c) 2019 University of Oxford
"""

from oxasl.gui.widgets import TabPage

class AslCalibration(TabPage):
    """
    Tab page containing options for calibration
    """

    def __init__(self, parent, idx, n):
        TabPage.__init__(self, parent, "Calibration", idx, n)

        self.calib_cb = self.checkbox("Enable Calibration", bold=True, handler=self.calib_changed)

        self.calib_image_picker = self.file_picker("Calibration Image")
        self.seq_tr_num = self.number("Sequence TR (s)", minval=0, maxval=10, initial=6)
        self.calib_gain_num = self.number("Calibration Gain", minval=0, maxval=5, initial=1)
        self.calib_mode_ch = self.choice("Calibration mode", choices=["Reference Region", "Voxelwise"])

        self.section("Reference tissue")

        self.ref_tissue_type_ch = self.choice("Type", choices=["CSF", "WM", "GM", "None"],
                                              handler=self.ref_tissue_type_changed)
        self.ref_tissue_mask_picker = self.file_picker("Mask", optional=True)
        self.ref_t1_num = self.number("Reference T1 (s)", minval=0, maxval=5, initial=4.3)
        self.seq_te_num = self.number("Sequence TE (ms)", minval=0, maxval=30, initial=0)
        self.ref_t2_num = self.number("Reference T2 (ms)", minval=0, maxval=1000, initial=750, step=10)
        self.blood_t2_num = self.number("Blood T2 (ms)", minval=0, maxval=1000, initial=150, step=10)
        self.coil_image_picker = self.file_picker("Coil Sensitivity Image", optional=True)

        self.sizer.AddGrowableCol(2, 1)
        self.SetSizer(self.sizer)
        self.next_prev()

    def options(self):
        options = {}
        if self.calib():
            options.update({
                "calib" : self.image("Calibration data", self.calib_image_picker.GetPath()),
                "calib_gain" : self.calib_gain_num.GetValue(),
                "tr" : self.seq_tr_num.GetValue(),
            })

            if self.refregion():
                options.update({
                    "calib_method" : "refregion",
                    "tissref" : self.ref_tissue_type_ch.GetString(self.ref_tissue_type()),
                    "te" : self.seq_te_num.GetValue(),
                    "t1r" : self.ref_t1_num.GetValue(),
                    "t2r" : self.ref_t2_num.GetValue(),
                    "t2b" : self.blood_t2_num.GetValue(),
                })
                if self.ref_tissue_mask_picker.checkbox.IsChecked():
                    options["refmask"] = self.ref_tissue_mask_picker.GetPath()
            else:
                options["calib_method"] = "voxelwise"

            if self.coil_image_picker.checkbox.IsChecked():
                options["cref"] = self.image("Calibration reference data", self.coil_image_picker.GetPath())

        return options

    def calib(self):
        """
        :return: True if calibration is enabled
        """
        return self.calib_cb.IsChecked()

    def refregion(self):
        """
        :return: True if reference region calibration is selected
        """
        return self.calib_mode_ch.GetSelection() == 0

    def ref_tissue_type(self):
        """
        :return reference tissue type index
        """
        return self.ref_tissue_type_ch.GetSelection()

    def ref_tissue_type_changed(self, _):
        """
        Update reference tissue parameters to currently selected reference tissue type
        """
        if self.ref_tissue_type() == 0: # CSF
            self.ref_t1_num.SetValue(4.3)
            self.ref_t2_num.SetValue(750)
        elif self.ref_tissue_type() == 1: # WM
            self.ref_t1_num.SetValue(1.0)
            self.ref_t2_num.SetValue(50)
        elif self.ref_tissue_type() == 2: # GM
            self.ref_t1_num.SetValue(1.3)
            self.ref_t2_num.SetValue(100)
        self.update()

    def calib_changed(self, _):
        """
        Update option visibility when calibration is enabled/disabled
        """
        self.distcorr.calib_changed(self.calib())
        self.update()

    def update(self):
        enable = self.calib()
        self.seq_tr_num.Enable(enable)
        self.calib_image_picker.Enable(enable)
        self.calib_gain_num.Enable(enable)
        self.coil_image_picker.checkbox.Enable(enable)
        if self.analysis.white_paper():
            self.calib_mode_ch.SetSelection(1)
        self.calib_mode_ch.Enable(enable and not self.analysis.white_paper())
        self.ref_tissue_type_ch.Enable(enable and self.refregion())

        if self.ref_tissue_type() == 3:
            # Ref tissue = None - enforce mask
            self.ref_tissue_mask_picker.checkbox.Enable(False)
            self.ref_tissue_mask_picker.checkbox.SetValue(enable and self.refregion())
            self.ref_tissue_mask_picker.Enable(enable and self.refregion())
        else:
            self.ref_tissue_mask_picker.checkbox.Enable(enable and self.refregion())
        self.ref_tissue_mask_picker.Enable(enable and self.refregion() and self.ref_tissue_mask_picker.checkbox.IsChecked())

        self.coil_image_picker.checkbox.Enable(enable and self.refregion())
        self.coil_image_picker.Enable(enable and self.refregion() and self.coil_image_picker.checkbox.IsChecked())
        self.seq_te_num.Enable(enable and self.refregion())
        self.blood_t2_num.Enable(enable and self.refregion())
        self.ref_t1_num.Enable(enable and self.refregion())
        self.ref_t2_num.Enable(enable and self.refregion())
        TabPage.update(self)



"""
Tests for BASIL module

FIXME need more multi-step tests
FIXME need to test more error conditions
"""
import pytest
import numpy as np

from fsl.data.image import Image

from oxasl import AslImage, Workspace
import oxasl.basil as basil

pytest.skip("skipping basil tests as neeed rewrite", allow_module_level=True)

DEFAULTS = {
    "method" : "vb",
    "noise" : "white",
    "allow-bad-voxels" : True,
    "convergence" : "trialmode",
    "max-iterations" : 20,
    "max-trials" : 10,
    "model" : "aslrest",
    "disp" : "none",
    "exch" : "mix",
    "inctiss" : True,
    "incbat" : True,
    "infertiss" : True,
    "inferbat" : True,
}

def _get_defaults(img):
    options = dict(DEFAULTS)
    for idx, ti in enumerate(img.tis):
        options["ti%i" % (idx+1)] = ti
        options["rpt%i" % (idx+1)] = img.rpts[idx]
    return options

def _check_step(step, desc_text=None, options=None):
    print(step.options)
    if desc_text:
        assert(desc_text.lower().strip() in step.desc.lower())
    if options:
        for k, v in options.items():
            assert(step.options[k] == v)

def test_nodata():
    """
    Check we get an error if there is no data
    """
    wsp = Workspace()

    with pytest.raises(ValueError):
        steps = basil.basil_steps(wsp, None)

def test_infer_nothing():
    """
    Check we get an error if there is nothing to infer
    """
    d = np.random.rand(5, 5, 5, 6)
    img = AslImage(name="asldata", image=d, tis=[1.5], order="prt")
    wsp = Workspace()

    with pytest.raises(ValueError):
        steps = basil.basil_steps(wsp, img)

def test_defaults():
    """
    Check the basic defaults (infer tissue perfusion and bolus arrival time)
    """
    d = np.random.rand(5, 5, 5, 6)
    img = AslImage(name="asldata", image=d, tis=[1.5], order="prt")
    wsp = Workspace()
    wsp.infertiss = True
    wsp.inferbat = True

    steps = basil.basil_steps(wsp, img)
    assert(len(steps) == 1)

    options = _get_defaults(img)
    _check_step(steps[0], desc_text="tissue", options=options)
                
def test_fix_bat():
    """
    Check fixing the arrival time, which is normally inferred
    """
    d = np.random.rand(5, 5, 5, 6)
    img = AslImage(name="asldata", image=d, tis=[1.5], order="prt")
    wsp = Workspace(infertiss=True, inferbat=False)

    steps = basil.basil_steps(wsp, img)
    assert(len(steps) == 1)

    options = _get_defaults(img)
    options.pop("incbat")
    options.pop("inferbat")
    _check_step(steps[0], desc_text="tissue", options=options)
 
def test_inferart():
    """
    Check inference of arterial component
    """
    d = np.random.rand(5, 5, 5, 6)
    img = AslImage(name="asldata", image=d, tis=[1.5], order="prt")
    wsp = Workspace(infertiss=True, inferbat=True, inferart=True)

    options = _get_defaults(img)
    steps = basil.basil_steps(wsp, img)
    assert(len(steps) == 2)

    _check_step(steps[0], desc_text="tissue", 
                options=dict(options, **{
                    "incart" : True,
                }))

    _check_step(steps[1], desc_text="arterial", 
                options=dict(options, **{
                    "incart" : True,
                    "inferart" : True,
                }))

def test_infertau():
    """
    Check inference of bolus duration (tau)
    """
    d = np.random.rand(5, 5, 5, 6)
    img = AslImage(name="asldata", image=d, tis=[1.5], order="prt")
    wsp = Workspace(infertiss=True, inferbat=True, infertau=True)

    steps = basil.basil_steps(wsp, img)
    assert(len(steps) == 2)

    options = _get_defaults(img)
    _check_step(steps[0], desc_text="tissue", 
                options=dict(options, **{
                    "inctau" : True,
                }))

    _check_step(steps[1], desc_text="bolus", 
                options=dict(options, **{
                    "inctau" : True,
                    "infertau" : True,
                }))

def test_inferarttau():
    """
    Check inference of bolus duration (tau) and arterial component
    """
    d = np.random.rand(5, 5, 5, 6)
    img = AslImage(name="asldata", image=d, tis=[1.5], order="prt")
    wsp = Workspace(infertiss=True, inferbat=True, infertau=True, inferart=True)

    steps = basil.basil_steps(wsp, img)
    assert(len(steps) == 3)

    options = _get_defaults(img)
    _check_step(steps[0], desc_text="tissue", 
                options=dict(options, **{
                    "inctau" : True,
                    "incart" : True,
                }))

    _check_step(steps[1], desc_text="arterial", 
                options=dict(options, **{
                    "inctau" : True,
                    "incart" : True,
                    "inferart" : True,
                }))

    _check_step(steps[2], desc_text="bolus", 
                options=dict(options, **{
                    "inctau" : True,
                    "incart" : True,
                    "infertau" : True,
                    "inferart" : True,
                }))

def test_infert1():
    """
    Check inference of T1
    """
    d = np.random.rand(5, 5, 5, 6)
    img = AslImage(name="asldata", image=d, tis=[1.5], order="prt")
    wsp = Workspace(infertiss=True, inferbat=True, infert1=True)

    steps = basil.basil_steps(wsp, img)
    assert(len(steps) == 2)

    options = _get_defaults(img)
    _check_step(steps[0], desc_text="tissue", 
                options=dict(options, **{
                    "inct1" : True,
                }))

    _check_step(steps[1], desc_text="T1", 
                options=dict(options, **{
                    "inct1" : True,
                    "infert1" : True,
                }))

def test_t1im():
    """
    Check T1 image priors are correctly handled
    """
    d = np.random.rand(5, 5, 5, 6)
    img = AslImage(name="asldata", image=d, tis=[1.5], order="prt")
    wsp = Workspace(infertiss=True, inferbat=True, infert1=True)

    wsp.t1im = Image(np.random.rand(5, 5, 5))
    steps = basil.basil_steps(wsp, img)
    assert(len(steps) == 2)

    options = _get_defaults(img)
    _check_step(steps[0], desc_text="tissue", 
                options=dict(options, **{
                    "inct1" : True,
                }))

    _check_step(steps[1], desc_text="T1", 
                options=dict(options, **{
                    "inct1" : True,
                    "infert1" : True,
                    "PSP_byname1" : "T_1",
                    "PSP_byname1_type" : "I",
                    #"PSP_byname1_image" : "t1file",
                }))

def test_inferpc():
    """
    Check the pre-capiliary component
    """
    d = np.random.rand(5, 5, 5, 6)
    img = AslImage(name="asldata", image=d, tis=[1.5], order="prt")
    wsp = Workspace(infertiss=True, inferbat=True, inferpc=True)

    steps = basil.basil_steps(wsp, img)
    assert(len(steps) == 2)

    options = _get_defaults(img)
    _check_step(steps[0], desc_text="tissue", 
                options=dict(options, **{
                    "incpc" : True,
                }))

    _check_step(steps[1], desc_text="pre-capiliary", 
                options=dict(options, **{
                    "incpc" : True,
                    "inferpc" : True,
                }))

def test_artonly():
    """
    Check we can infer arterial component without tissue step
    """
    d = np.random.rand(5, 5, 5, 6)
    img = AslImage(name="asldata", image=d, tis=[1.5], order="prt")
    wsp = Workspace(infertiss=False, inferbat=True, inferart=True)

    steps = basil.basil_steps(wsp, img)
    assert(len(steps) == 1)

    options = _get_defaults(img)
    options.update({
        "incart" : True,
        "inferart" : True,
    })
    options.pop("inctiss")
    options.pop("infertiss")
    _check_step(steps[0], desc_text="arterial", options=options)

def test_initmvn():
    """
    Check the supply of an initialization MVN
    """
    d = np.random.rand(5, 5, 5, 6)
    img = AslImage(name="asldata", image=d, tis=[1.5], order="prt")
    wsp = Workspace(infertiss=True, inferbat=True)

    wsp.initmvn = Image(np.random.rand(5, 5, 5, 6))
    steps = basil.basil_steps(wsp, img)
    assert(len(steps) == 1)

    options = _get_defaults(img)
    options.update({
        "continue-from-mvn" : wsp.initmvn
    })
    _check_step(steps[0], desc_text="tissue", options=options)

def test_spatial():
    """
    Check final spatial step
    """
    d = np.random.rand(5, 5, 5, 6)
    img = AslImage(name="asldata", image=d, tis=[1.5], order="prt")
    wsp = Workspace(infertiss=True, inferbat=True, spatial=True)

    steps = basil.basil_steps(wsp, img)
    assert(len(steps) == 2)
    
    options = _get_defaults(img)
    _check_step(steps[0], desc_text="tissue", options=options)

    options.update({
        "method" : "spatialvb",
        "param-spatial-priors" : "N+",
        "PSP_byname1" : "ftiss",
        "PSP_byname1_type" : "M",
        "convergence" : "maxits",
    })
    options.pop("max-trials")
    _check_step(steps[1], desc_text="spatial", options=options)
    
def test_onestep():
    """
    Check that single step mode works when you would normally get multiple steps
    """
    d = np.random.rand(5, 5, 5, 6)
    img = AslImage(name="asldata", image=d, tis=[1.5], order="prt")
    wsp = Workspace(infertiss=True, inferbat=True, infertau=True, inferart=True, spatial=True, onestep=True)

    steps = basil.basil_steps(wsp, img)
    assert(len(steps) == 1)

    options = _get_defaults(img)
    options.update({
        "method" : "spatialvb",
        "param-spatial-priors" : "N+",
        "PSP_byname1" : "ftiss",
        "PSP_byname1_type" : "M",
        "inctau" : True,
        "incart" : True,
        "inferart" : True,
        "infertau" : True,
        "convergence" : "maxits",
    })
    options.pop("max-trials")
    _check_step(steps[0], desc_text="spatial", options=options)

def test_max_iterations():
    """
    Check that max iterations can be overridden
    """
    d = np.random.rand(5, 5, 5, 6)
    img = AslImage(name="asldata", image=d, tis=[1.5], order="prt")
    wsp = Workspace(infertiss=True, inferbat=True)

    kwargs = {
        "max-iterations" : 123,
    }
    steps = basil.basil_steps(wsp, img, **kwargs)
    assert(len(steps) == 1)

    options = _get_defaults(img)
    options.update({
        "max-iterations" : 123,
    })
    _check_step(steps[0], desc_text="tissue", options=options)

def test_random_extra_options():
    """
    Check that any additional keyword arguments are passed to Fabber
    """
    d = np.random.rand(5, 5, 5, 6)
    img = AslImage(name="asldata", image=d, tis=[1.5], order="prt")
    wsp = Workspace(infertiss=True, inferbat=True)

    kwargs = {
        "phase-of-moon-correction-factor" : 7,
        "random-output-proportion-percent" : 36,
    }
    steps = basil.basil_steps(wsp, img, **kwargs)
    assert(len(steps) == 1)

    options = _get_defaults(img)
    options.update(kwargs)
    _check_step(steps[0], desc_text="tissue", options=options)

def test_pvc_only_one_map_given1():
    """
    Check that PVC correction fails if you only give the GM map
    """
    d = np.random.rand(5, 5, 5, 6)
    img = AslImage(name="asldata", image=d, tis=[1.5], order="prt")
    wsp = Workspace(infertiss=True, inferbat=True)

    wsp.pgm = Image(np.random.rand(5, 5, 5))
    with pytest.raises(ValueError):
        basil.basil_steps(wsp, img)
    
def test_pvc_only_one_map_given2():
    """
    Check that PVC correction fails if you only give the WM map
    """
    d = np.random.rand(5, 5, 5, 6)
    img = AslImage(name="asldata", image=d, tis=[1.5], order="prt")
    wsp = Workspace(infertiss=True, inferbat=True)

    wsp.pwm = Image(np.random.rand(5, 5, 5))
    with pytest.raises(ValueError):
        basil.basil_steps(wsp, img)
    
def test_pvc_no_tissue():
    """
    Check that PVC correction fails if you do not infer the tissue component
    """
    d = np.random.rand(5, 5, 5, 6)
    img = AslImage(name="asldata", image=d, tis=[1.5], order="prt")
    wsp = Workspace(infertiss=False, inferbat=True)

    wsp.pgm = Image(np.random.rand(5, 5, 5))
    wsp.pwm = Image(np.random.rand(5, 5, 5))

    with pytest.raises(ValueError):
        basil.basil_steps(wsp, img)
    
def test_pvc():
    """
    FIXME we need to test the PVC initialization step
    and how to do this is not finalized
    """
    d = np.random.rand(5, 5, 5, 6)
    img = AslImage(name="asldata", image=d, tis=[1.5], order="prt")
    wsp = Workspace(infertiss=True, inferbat=True)

    wsp.pgm = Image(np.random.rand(5, 5, 5))
    wsp.pwm = Image(np.random.rand(5, 5, 5))
    
    steps = basil.basil_steps(wsp, img)
    assert(len(steps) == 3)

    options = _get_defaults(img)
#    options.update({
#        "incpve" : True,
#    })
    _check_step(steps[0], desc_text="tissue", options=options)
#    _check_step(steps[1], desc_text="PVE", options=options)

    options.update({
        "method" : "spatialvb",
        "param-spatial-priors" : "N+",
        "PSP_byname1" : "ftiss",
        "PSP_byname1_type" : "M",
        "max-iterations" : 200,
        "convergence" : "maxits",
    })
    options.pop("max-trials")
    _check_step(steps[2], desc_text="spatial")




"""
oxasl.gui.analysis_tab.py

Copyright (c) 2019 University of Oxford
"""
import os

from oxasl.gui.widgets import TabPage, OptionError

class AslAnalysis(TabPage):
    """
    Tab page containing model fitting and analysis options
    """

    def __init__(self, parent, idx, n):
        TabPage.__init__(self, parent, "Analysis", idx, n)

        self.distcorr_choices = ["Fieldmap", "Calibration image"]

        self.section("Basic analysis options")

        self.outdir_picker = self.file_picker("Output Directory", pick_dir=True)
        self.mask_picker = self.file_picker("Brain Mask", optional=True)
        self.wp_cb = self.checkbox("Analysis which conforms to 'White Paper' (Alsop et al 2014)",
                                   handler=self.wp_changed)

        self.section("Initial parameter values")

        self.bat_num = self.number("Arterial Transit Time (s)", minval=0, maxval=2.5, initial=1.3)
        self.t1_num = self.number("T1 (s)", minval=0, maxval=3, initial=1.3)
        self.t1b_num = self.number("T1b (s)", minval=0, maxval=3, initial=1.65)
        self.ie_num = self.number("Inversion Efficiency", minval=0, maxval=1, initial=0.85)

        self.section("Analysis Options")

        self.spatial_cb = self.checkbox("Adaptive spatial regularization on perfusion", initial=True)
        self.infer_t1_cb = self.checkbox("Incorporate T1 value uncertainty")
        self.macro_cb = self.checkbox("Include macro vascular component")
        self.fixbolus_cb = self.checkbox("Fix label duration", initial=True)

        self.pv_cb = self.checkbox("Partial Volume Correction")
        self.mc_cb = self.checkbox("Motion Correction")

        self.sizer.AddGrowableCol(1, 1)
        #sizer.AddGrowableRow(5, 1)
        self.SetSizer(self.sizer)
        self.next_prev()

    def options(self):
        outdir = self.outdir_picker.GetPath()
        if not outdir:
            raise OptionError("Output directory must be specified")
        elif os.path.isfile(outdir):
            raise OptionError("Output directory already exists and is a file")

        options = {
            "savedir" : outdir,
            "wp" : self.white_paper(),
            "bat" : self.bat_num.GetValue(),
            "t1" : self.t1_num.GetValue(),
            "t1b" : self.t1b_num.GetValue(),
            "alpha" : self.ie_num.GetValue(),
            "spatial" : self.spatial_cb.IsChecked(),
            "infert1" : self.infer_t1_cb.IsChecked(),
            "inferart" : self.macro_cb.IsChecked(),
            "fixbolus" : self.fixbolus_cb.IsChecked(),
            "pvcorr" : self.pv_cb.IsChecked(),
            "mc" : self.mc_cb.IsChecked(),
        }

        if self.mask_picker.checkbox.IsChecked():
            options["mask"] = self.mask_picker.GetPath()

        return options

    def white_paper(self):
        """
        :return: True if white paper mode is enabled
        """
        return self.wp_cb.IsChecked()

    def update(self):
        self.mask_picker.Enable(self.mask_picker.checkbox.IsChecked())
        self.t1_num.Enable(not self.white_paper())
        self.bat_num.Enable(not self.white_paper())
        TabPage.update(self)

    def wp_changed(self, _):
        """
        Update values to white paper defaults when WP mode selected
        and our defaults when it is deselected
        """
        if self.white_paper():
            self.t1_num.SetValue(1.65)
            self.bat_num.SetValue(0)
        else:
            self.t1_num.SetValue(1.3)
            self.bat_num.SetValue(1.3)
        self.calibration.update()
        self.update()

    def labelling_changed(self, pasl):
        """
        Update values to PASL/pCASL defaults when labelling method changed
        """
        if pasl:
            self.bat_num.SetValue(0.7)
            self.ie_num.SetValue(0.98)
        else:
            self.bat_num.SetValue(1.3)
            self.ie_num.SetValue(0.85)



"""
Classes for parsing command line options for ASL tools

Copyright (c) 2008-2018 University of Oxford
"""
import sys
from optparse import OptionGroup, OptionParser, Option, OptionValueError
from collections import defaultdict
from copy import copy
import tempfile

import numpy as np

from fsl.data.image import Image

from oxasl import __version__

class AslOptionParser(OptionParser):
    """
    OptionParser which is extended to include the concept of option categories

    A category contains one or more OptionGroup instances. An option
    dictionary can be filtered by category, i.e. to extract the options
    belonging to a particular category (e.g. 'calibration options'). The
    existing OptionGroup class is not sufficient for this because groups
    cannot be nested
    """
    def __init__(self, usage="", version=__version__, **kwargs):
        OptionParser.__init__(self, usage=usage, version=version, option_class=_ImageOption, **kwargs)
        self._categories = defaultdict(list)

    def parse_args(self, argv=None, values=None):
        if argv is None:
            argv = sys.argv[1:]
        options, args = OptionParser.parse_args(self, argv, values)
        if options.optfile:
            # When an option file is specifeid, extract the options, build
            # a new argv vector and re-parse it. This is the only way to ensure
            # that options in the file work identically to CLI options.
            new_argv = self._add_from_file(argv, options.optfile)
            options, args = OptionParser.parse_args(self, new_argv, values)

        # Deal with case where asldata is given as separate files
        if args and options.asldata is None:
            merged_data = None
            for idx, fname in enumerate(args):
                img = Image(fname)
                shape = list(img.shape)
                if img.ndim == 3:
                    shape += [1,]
                if merged_data is None:
                    merged_data = np.zeros(shape[:3] + [shape[3] * len(args)])
                merged_data[..., idx*shape[3]:(idx+1)*shape[3]] = img.data
            merged_img = Image(merged_data, header=img.header)
            temp_asldata = tempfile.NamedTemporaryFile(prefix="oxasl", delete=True)
            options.asldata = temp_asldata.name
            merged_img.save(options.asldata)

        return options, args

    def _add_from_file(self, argv, optfile):
        new_argv = list(argv)
        with open(optfile, "r") as f:
            optlines = f.readlines()

        for line in optlines:
            line = line[:line.find("#")].strip()
            if line:
                line = line.lstrip("-").replace(":", " ").replace("=", " ")
                kv = line.split(None, 1)
                key = "-" + kv[0]
                if len(kv[0]) > 1:
                    key = "-" + key
                new_argv.append(key)
                if len(kv) == 2:
                    new_argv.append(kv[1])
        return new_argv

    def add_category(self, category):
        """
        Add an OptionCategory to the parser
        """
        for group in category.groups(self):
            self.add_option_group(group)
            for option in group.option_list:
                self._categories[category.name].append(option)

class OptionCategory(object):
    """
    A named category of options.
    """
    def __init__(self, name):
        self.name = name

    def groups(self, parser):
        """
        :param parser: OptionParser instance
        :return: Sequence of OptionGroup instances for this category of options
        """
        return []

class GenericOptions(OptionCategory):
    """
    OptionCategory which contains generic options common to many command line tools
    """

    def __init__(self, title="Generic", output_type="directory", default_output="oxasl", **kwargs):
        OptionCategory.__init__(self, "generic", **kwargs)
        self.title = title
        self.output_type = output_type
        self.default_output = default_output

    def groups(self, parser):
        group = OptionGroup(parser, self.title)
        group.add_option("--output", "-o", help="Output %s" % self.output_type, default=self.default_output)
        group.add_option("--overwrite", help="Overwrite output %s if it already exists" % self.output_type, action="store_true", default=False)
        group.add_option("--mask", "-m", help="Brain mask image in ASL space", default=None, type="image")
        group.add_option("--optfile", help="File containing additional options")
        group.add_option("--log-cmds", help="Log all external commands run", action="store_true", default=False)
        group.add_option("--log-cmdout", help="Log the standard output of all external commands run", action="store_true", default=False)
        group.add_option("--debug", help="Debug mode - log all command output and keep all output files", action="store_true", default=False)
        return [group, ]

def load_options_file(fname):
    options = {}
    if fname:
        with open(fname) as options_file:
            for line in options_file:
                keyval = line.strip().rstrip("\n").lstrip("--").split("=", 1)
                key = keyval[0].strip()
                if key != "":
                    if len(keyval) == 1:
                        options[key] = True
                    else:
                        options[key] = keyval[1].strip()
    return options

def load_matrix(fname):
    """
    Load a matrix from a text file

    Matrix should be a series of ASCII floating point numbers separated by whitespace
    on each line.

    :param fname: Filename
    """
    matrix = []
    with open(fname, "r") as f:
        for line in f.readlines():
            if line.strip():
                matrix.append([float(v) for v in line.strip().split()])
    return np.array(matrix, dtype=np.float32)

def _check_image(option, opt, value):
    try:
        return Image(value, loadData=False)
    except ValueError:
        raise OptionValueError("option %s: invalid Image value: %r" % (opt, value))

def _check_matrix(option, opt, value):
    try:
        return load_matrix(value)
    except ValueError:
        raise OptionValueError("option %s: invalid matrix value: %r" % (opt, value))

def _check_optfile(option, opt, value):
    try:
        return load_options_file(value)
    except ValueError:
        raise OptionValueError("option %s: invalid options file: %r" % (opt, value))

class _ImageOption(Option):
    TYPES = Option.TYPES + ("image", "matrix", "optfile",)
    TYPE_CHECKER = copy(Option.TYPE_CHECKER)
    TYPE_CHECKER["image"] = _check_image
    TYPE_CHECKER["matrix"] = _check_matrix
    TYPE_CHECKER["optfile"] = _check_optfile



#!/bin/env python
"""
OXASL - Prefiltering module

This is run before any model fitting and is intended for volume selection, denoising etc.
Pre-filtering processes do not depend on the type of ASL data (e.g. multiphase, vessel-encoded)

Copyright (c) 2008-2020 Univerisity of Oxford
"""

try:
    import oxasl_enable
except ImportError:
    oxasl_enable = None

try:
    import oxasl_deblur
except ImportError:
    oxasl_deblur = None

def run(wsp):
    wsp.sub("filter")

    if oxasl_enable and wsp.use_enable:
        oxasl_enable.run(wsp, output_wsp=wsp.filter)

    if oxasl_deblur and wsp.deblur:
        oxasl_deblur.run(wsp, output_wsp=wsp.filter)
